/** \file exports.h
	\brief Defines the exported functions and structures
*/

#ifndef AUDRECORDLIB_EXPORTS_H
#define AUDRECORDLIB_EXPORTS_H

#pragma once

/** \def AUDRECORD_API 
	Resolves to nothing if building the AudRecord dll
	otherwise to __declspec(dllimport)
*/
#ifdef AUDRECORD_EXPORTS
#define AUDRECORD_API
#else
#define AUDRECORD_API __declspec(dllimport)
#endif

#include <wmsdk.h>
#include <mmdeviceapi.h>
#include "IAudRecordInterface.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Information regarding a single codec configuration */
typedef struct _CodecConfiguration
{
	/** A description of this configuration. 
		For example: "128 kbps, 48 kHz, 2 channel 24 bit CBR"
	*/
	LPCWSTR pName;
	/** The configuration interface pointer.
	
		The interface representing the configuration.
		This can be used to get more advanced information
		and the exact format of the output samples by QueryInterface-ing
		for IWMInputMediaProps.
		Do not call Release on this pointer.
	*/
	IWMStreamConfig* pConfig;
} CodecConfiguration;

/** Information about an installed codec */
typedef struct _Codec
{
	/** The name of the codec. 
		For example: "Windows Media Audio 10 Professional
	*/
	LPCWSTR name;
	/** The number of configurations in the configs array
	*/
	DWORD numConfigs;
	/** An array of compression configurations. One for each
		type the codec supports
	*/
	CodecConfiguration configs[1];
} Codec;

/** A collection of codecs and their configurations */
typedef struct _CodecConfigs
{
	/** The number of Codec pointers in the pCodecs array */
	DWORD num;
	/** The array of available codecs */
	Codec* pCodecs[1];
} CodecConfigs;

/** Represents a single audio rendering device. */
typedef struct _Endpoint
{
	/** The name/description of the device.
		For example: "Speakers (High Definition Audio Device)"
	*/
	LPCWSTR pName;
	/** The interface representing the audio device
		Do not call Release on this pointer.
	*/
	IMMDevice* pEndPoint;
}Endpoint;

/** A collection of audio render devices which can potentially be captured from */
typedef struct _EndpointArray
{
	/** The number of endpoints in the outputs array */
	DWORD num;
	/** The array of audio capture endpoints */
	Endpoint outputs[1];
}EndpointArray;

/** Returns an list of all installed audio codec configurations.

	Retrieves the list of usable codecs and a list of each CBR and VBR configuration they support.
	These configurations can be passed to CreateAudRecordInterface and IAudRecord::ChangeCodec to 
	define the compression format and bitrate used.
	CoInitialize must be called prior to calling this function
	\param [out] ppConfigs A pointer to a pointer that receives the allocated codec configurations
	\return S_OK on success, standard COM error code on failure including E_POINTER and CO_E_NOTINITIALIZED
	\sa CodecConfigs, FreeWMACodecConfigs, CreateAudRecordInterface and IAudRecord::ChangeCodec
*/

AUDRECORD_API STDMETHODIMP GetWMACodecConfigs(CodecConfigs** ppConfigs);

/** Frees a list of codec configurations allocated by GetWMACodecConfigs

	\param pConfigs The allocated pointer retrieved from GetWMACodecConfigs
	\return Nothing
	\sa GetWMACodecConfigs
*/
AUDRECORD_API STDMETHODIMP_(void) FreeWMACodecConfigs(CodecConfigs* pConfigs);

/** Returns an list of all installed audio output devices.

	Retrieves the list of possible devices to capture audio on along with their name and
	description. One of these endpoints must be passed to IAudRecord::Record to begin recording.
	CoInitialize must be called prior to calling this function and the returned pointer must be
	freed with FreeAudioEndpoints when it is no longer required
	\param [out] ppEndpoints A pointer to a pointer that receives the allocated audio endpoints
	\return S_OK on success, standard COM error code on failure including E_POINTER and CO_E_NOTINITIALIZED
	\sa EndpointArray, FreeAudioEndpoints, IAudRecord::Record
*/
AUDRECORD_API STDMETHODIMP GetAudioEndpoints(EndpointArray** ppEndpoints);

/** Frees a list of audio render endpoints allocated by GetAudioEndpoints

	\param pEndpoints The allocated pointer retrieved from GetAudioEndpoints
	\sa GetAudioEndpoints
*/
AUDRECORD_API STDMETHODIMP_(void) FreeAudioEndpoints(EndpointArray* pEndpoints);

/** Create an IAudRecord interface using the specified codec.

	Creates a new IAudRecord interface and calls ChangeCodec with the pCodecConfig
	parameter. Recording does not start until you add one or more targets with the
	AddFileTarget, AddNetworkTarget, or AddCustomTarget member functions and call Record
	with an Endpoint returnen from GetAudioEndpoints.
	COM must be initialized before calling this function. When you no longer need the IAudRecord
	object, free it by calling Release()
	\param pCodecConfig The initial codec configuration to use as for the output
	\param [out] ppAudRecord
	\return S_OK on success, standard COM error on failure including E_POINTER and CO_E_NOTINITIALIZED
	\sa CodecConfiguration, IAudRecord
*/
AUDRECORD_API STDMETHODIMP CreateAudRecordInterface(CodecConfiguration* pCodecConfig, IAudRecord** ppAudRecord);


/** Typedef for the custom alloc function form */
typedef PVOID (WINAPI*pfnAlloc)(PVOID pUserData, SIZE_T bytesRequired);
/** Custom free function form */
typedef void (WINAPI*pfnFree)(PVOID pUserData, PVOID pMemory);
/** Sets the functions used to allocate and free the memory AudRecordLib requires

	If this function isn't called, HeapAlloc and HeapFree are used.
	\param allocFn Function which should allocate memory
	\param freeFn Frees the memory returned by allocFn
	\param pUserData Opaque blob of context data that is passed to both functions untouched
	\returns The value of pUserData from the previous call
*/
AUDRECORD_API STDMETHODIMP_(PVOID) SetAllocFreeFunctions(pfnAlloc allocFn, pfnFree freeFn, PVOID pUserData);

/** In debug builds, AudRecord keeps a list of all allocations made
	either by the default allocation functions or by those set by
	SetAllocFreeFunctions(). Calling this function dumps all 
	allocations and stack traces via both the ETW debug
	stream and the debuggers text output via OutputDebugStringA()
*/
AUDRECORD_API STDMETHODIMP_(void) DumpUnfreedAllocations();

#ifdef __cplusplus
} // end of extern C
#endif

#endif
