/** \file AudRecordWrap.cpp
	\brief Implements the wrapper of the native IAudRecord interface
*/

#include "stdafx.h"
#include "AudRecordWrap.h"
#include "error.h"
#include "InterfaceWrap.h"
#include <exception>

using namespace System::Runtime::InteropServices;

/** \def NATIVE_ENDPOINT_FROM_MANAGED(managed, native)
	Fills in a native Endpoint struct using the info in a managed
	AudRecordCLI::Endpoint struct. This isn't a function because 
	the pin_ptr needs to live for the lifetime of the cstring
	it produces
*/
#define NATIVE_ENDPOINT_FROM_MANAGED(managed, native) \
	pin_ptr<const WCHAR> _name_string = PtrToStringChars(managed->Name); \
	{ \
		IntPtr origUnknownPtr = Marshal::GetIUnknownForObject(managed->Interface); \
		IMMDevice* pMMDevice = (static_cast<IMMDevice*>(origUnknownPtr.ToPointer())); \
		/* We don't need a reference to persist, since managed->Interface */ \
		/* already has a reference that will outlive the call */ \
		Marshal::Release(origUnknownPtr); \
		IMMDevice* pTest = NULL; \
		HRESULT hrQI = S_OK; \
		/* assert that the returned interface of the correct type */ \
		ASSERT((hrQI = pMMDevice->QueryInterface(&pTest), (pTest ? pTest->Release() : 0ul), SUCCEEDED(hrQI))); \
		native.pEndPoint = pMMDevice; \
		native.pName = _name_string; \
	}

/** \def NATIVE_CODECCONFIG_FROM_MANAGED(managed, native)
	Fills in a native CondecConfig struct using the info in a managed
	AudRecordCLI::CodecConfiguration struct. This isn't a function because 
	the pin_ptr needs to live for the lifetime of the cstring
	it produces
*/
#define NATIVE_CODECCONFIG_FROM_MANAGED(managed, native) \
	pin_ptr<const WCHAR> _name_string = PtrToStringChars(managed->Name); \
	{ \
		IntPtr origUnknownPtr = Marshal::GetIUnknownForObject(managed->DescInterface); \
		IWMStreamConfig* pConfig = (static_cast<IWMStreamConfig*>(origUnknownPtr.ToPointer())); \
		/* We don't need a reference to persist, since managed->Interface */ \
		/* already has a reference that will outlive the call */ \
		Marshal::Release(origUnknownPtr); \
		IWMStreamConfig* pTest = NULL; \
		HRESULT hrQI = S_OK; \
		/* assert that the returned interface of the correct type */ \
		ASSERT((hrQI = pConfig->QueryInterface(&pTest), (pTest ? pTest->Release() : 0ul), SUCCEEDED(hrQI))); \
		native.pConfig = pConfig; \
		native.pName = _name_string; \
	}

/** 
*/
namespace AudRecordCLI
{
	/** Construct a Recorder which controls a specified
		underlying IAudRecord interface

		\param pRecord The interface to wrap
	*/
	Recorder::Recorder(IAudRecord* pRecord)
		: pNativeRecord(pRecord),
		wrapObj(Marshal::GetObjectForIUnknown(IntPtr(pRecord)))
	{
		// the GetObjectForIUnknown above increases the ref count,
		// so this releases it. This means the object will be fully released
		// when wrapObj is destructed. As it's managed too, it means we don't
		// have to bother with IDispose and all that jazz
		pNativeRecord->Release();
	}

	/** Starts the recording process.

		Reconstructs a native Endpoint struct from theEndpoint and passes
		that to the native Record function
		\exception ArgumentNullException if theEndpoint is null, or an
		AudRecordException, OutOfMemoryException or ArgumentException based
		on the return on the native function
		\param theEndpoint The mixer to record from
	*/
	void Recorder::Record(Endpoint^ theEndpoint)
	{
		THROW_ON_NULL_ARG(theEndpoint, NULL_ENDPOINT_ERR);
		// reconstruct a native endpoint
		::Endpoint endPoint;
		NATIVE_ENDPOINT_FROM_MANAGED(theEndpoint, endPoint);
		HRESULT hr = pNativeRecord->Record(&endPoint);
		ThrowExceptionIfRequired(hr);
	}

	/** Changes the stream codec or configuration

		Turns a managed CodecConfig into a native CodecConfiguration struct
		and calls the native ChangeCodec function
		\exception ArgumentNullException if theEndpoint is null, or an
		AudRecordException, OutOfMemoryException or ArgumentException based
		on the return on the native function
		\param theConfig The new codec and configuration
	*/
	void Recorder::ChangeCodec(CodecConfig^ theConfig)
	{
		THROW_ON_NULL_ARG(theConfig, NULL_CODEC_ERR);
		// reconstruct a native codec
		::CodecConfiguration codecConf;
		NATIVE_CODECCONFIG_FROM_MANAGED(theConfig, codecConf);
		HRESULT hr = pNativeRecord->ChangeCodec(&codecConf);
		ThrowExceptionIfRequired(hr);
	}

	/** Pauses/Resumes the recording stream

		\exception An AudRecordException, OutOfMemoryException or ArgumentException
		based on the return on the native function
		\returns The state of the stream after the toggle operation
	*/
	AudPausedState Recorder::TogglePause()
	{
		::AudPausedState state = ::AudPaused;
		HRESULT hr = pNativeRecord->TogglePause(&state);
		ThrowExceptionIfRequired(hr);
		return (state == ::AudPaused) ? AudPausedState::AudPaused : AudPausedState::AudRecording;
	}

	/** Queries the current state of the stream

		\exception An AudRecordException, OutOfMemoryException or ArgumentException
		based on the return on the native function
		\returns The state of the stream after the toggle operation
	*/
	AudPausedState Recorder::QueryPause()
	{
		::AudPausedState state = ::AudPaused;
		HRESULT hr = pNativeRecord->QueryPause(&state);
		ThrowExceptionIfRequired(hr);
		return (state == ::AudPaused) ? AudPausedState::AudPaused : AudPausedState::AudRecording;
	}

	/** Completely stops the recording process.
		Record() must be called to restart it

		\exception An AudRecordException, OutOfMemoryException or ArgumentException 
		based on the return on the native function
	*/
	void Recorder::Stop()
	{
		HRESULT hr = pNativeRecord->Stop();
		ThrowExceptionIfRequired(hr);
	}

	/**  Adds a file to the recording process, can be called at any time

		\param fileName The file to create
		\exception ArgumentNullException if fileName is null, or an
		AudRecordException, OutOfMemoryException or ArgumentException based
		on the return on the native function
		\returns An opaque value that can be used to remove the stream
		at a later time
	*/
	IntPtr Recorder::AddFileTarget(String^ fileName)
	{
		THROW_ON_NULL_ARG(fileName, NULL_FILENAME_ERR);
		pin_ptr<const WCHAR> stringChars = PtrToStringChars(fileName);
		PVOID pCookie = NULL;
		HRESULT hr = pNativeRecord->AddFileTarget(stringChars, &pCookie);
		ThrowExceptionIfRequired(hr);
		return IntPtr(pCookie);
	}

	/** Opens a network port and streams the recording through it, can be called at any time

		\exception An AudRecordException, OutOfMemoryException or ArgumentException 
		based on the return on the native function
		\param port The port to use, or 0 for a randomly selected one
		\param maxConnections The number of people who can connect to the stream. 5 is the default
		\param wantUrl Whether to return the url to connect to in the url parameter
		\param [out] url The connection url, only returned if wantUrl is true
		\returns An opaque value used to remove the target 
	*/
	IntPtr Recorder::AddNetworkBroadcastTarget(DWORD% port, DWORD maxConnections, bool wantUrl, String^% url)
	{
		DWORD nativePortVar = 0;
		DWORD* pNativePortToUse = &nativePortVar;
		// pin the port if it was specified
		pin_ptr<DWORD> portPtr;
		if(port)
		{
			portPtr = interior_ptr<DWORD>(&port);
			pNativePortToUse = portPtr;
		}
		// only get the url string if they wanted it
		LPWSTR pNativeUrl = NULL;
		LPWSTR* ppNativeUrl = NULL;
		if(wantUrl)
		{
			ppNativeUrl = &pNativeUrl;
		}
		PVOID pCookie = NULL;
		HRESULT hr = pNativeRecord->AddNetworkBroadcastTarget(pNativePortToUse, maxConnections, ppNativeUrl, &pCookie);
		ThrowExceptionIfRequired(hr);
		if(pNativeUrl)
		{
			url = gcnew String(pNativeUrl);
			CoTaskMemFree(pNativeUrl);
		}
		return IntPtr(pCookie);
	}

	/** Adds a target that can manipulate the output audio how it wants,
		can be called at any time

		\exception ArgumentNullException if pTarget is null, or an
		AudRecordException, OutOfMemoryException or ArgumentException based
		on the return on the native function
		\param pTarget The object that receives the audio data
		and related notifications
		\returns An opaque value used to remove the target 
	*/
	IntPtr Recorder::AddCustomTarget(ITargetSink^ pTarget)
	{
		THROW_ON_NULL_ARG(pTarget, NULL_INTERFACE_ERR);
		ComType<TargetSinkWrap> pWrap;
		HRESULT hr = S_OK;
		try
		{
			pWrap.Init(new TargetSinkWrap(pTarget));
		}
		catch(const std::exception& ex)
		{
			hr = E_OUTOFMEMORY;
			DBG_TEXT("Caught exception %s creating a TargetSinkWrap", ex.what());
		}
		PVOID pCookie = NULL;
		if(SUCCEEDED(hr))
		{
			hr = pNativeRecord->AddCustomTarget(*pWrap, &pCookie);
		}
		ThrowExceptionIfRequired(hr);
		return IntPtr(pCookie);
	}

	/** Removes a target from the recording process, can be called at any time

		\exception An AudRecordException, OutOfMemoryException or ArgumentException 
		based on the return on the native function
		\param pCookie The value returned by AddCustomTarget(),
		AddNetworkBroadcastTarget() or AddFileTarget()
	*/
	void Recorder::RemoveTarget(IntPtr pCookie)
	{
		HRESULT hr = pNativeRecord->RemoveTarget(pCookie.ToPointer());
		ThrowExceptionIfRequired(hr);
	}

	/** Returns current statistics from the recording process

		\exception An AudRecordException, OutOfMemoryException or ArgumentException 
		based on the return on the native function
		\returns The structure containing the statistics
	*/
	AudRecordStats^ Recorder::GetStatistics()
	{
		Type^ audStatsType = AudRecordStats::typeid;
		ASSERT(Marshal::SizeOf(audStatsType) == sizeof(::AudRecordStats));

		::AudRecordStats nativeStats;
		HRESULT hr = pNativeRecord->GetStatistics(&nativeStats);
		ThrowExceptionIfRequired(hr);
		return static_cast<AudRecordStats^>(Marshal::PtrToStructure(IntPtr(&nativeStats), audStatsType));
	}

	/** Creates a new recorder initialised with a specific configuration

		\exception ArgumentNullException if initialConfig is null, or an
		AudRecordException, OutOfMemoryException or ArgumentException based
		on the return on the native function
		\param initialConfig Codec and bitrate to start the recorder off with 
		\returns The new wrapper object
	*/
	Recorder^ Recorder::Create(CodecConfig^ initialConfig)
	{
		THROW_ON_NULL_ARG(initialConfig, NULL_CODEC_ERR);
		// reconstruct a native codec
		::CodecConfiguration codecConf;
		NATIVE_CODECCONFIG_FROM_MANAGED(initialConfig, codecConf);
		IAudRecord* pNativeRecord = NULL;
		HRESULT hr = CreateAudRecordInterface(&codecConf, &pNativeRecord);
		ThrowExceptionIfRequired(hr);
		return gcnew Recorder(pNativeRecord);
	}
}
