/** \file DataBufferWrap.cpp
	\brief Implements the managed wrapper around the INSSBuffer interface
	and th native wrapper around 
	\cond DEV
*/

#include "stdafx.h"
#include "DataBufferWrap.h"
#include <exception>

using namespace System::Runtime::InteropServices;

/** Constructs the wrapper around the specified INSSBuffer

	\param pBuffer The buffer to wrap
*/
NativeDataBufferWrap::NativeDataBufferWrap(INSSBuffer* pBuffer)
	: pNSBuffer(pBuffer),
	interfaceWrap(Marshal::GetObjectForIUnknown(IntPtr(pBuffer)))
{
	// GetObjectForIUnknown raises the refcount
	// as we have exclusive access, we release to return
	// the count to one, while retianing the pointer for
	// easy access
	pBuffer->Release();
}

/** Wraps a native pointer returned by the native method of the same name
	in an IntPtr for easier managed consumption

	\returns An IntPtr holding the address of an allocated buffer
*/
IntPtr NativeDataBufferWrap::GetBuffer()
{
	BYTE* pBuffer = NULL;
	DWORD length = 0;
	HRESULT hr = pNSBuffer->GetBuffer(&pBuffer);
	if(FAILED(hr))
	{
		DBG_TEXT("GetBuffer failed with error %#08x", hr);
		return IntPtr::Zero;
	}
	return IntPtr(pBuffer);
}

/** Does a job similar to GetBuffer with the addition of returning
	the length parameter.

	\param [out] length The size of the allocated buffer pointed to by the 
	returned IntPtr
	\returns An IntPtr holding the address of an allocated buffer
*/
IntPtr NativeDataBufferWrap::GetBufferAndLength([Out] DWORD %length)
{
	pin_ptr<DWORD> pLength = interior_ptr<DWORD>(&length);
	HRESULT hr = pNSBuffer->GetLength(pLength);
	if(FAILED(hr))
	{
		DBG_TEXT("GetLength failed with error %#08x", hr);
		return IntPtr::Zero;
	}
	return GetBuffer();
}

/** Queries the current number of bytes used in the buffer,
	set by a previous call to SetLength

	\returns Number of bytes used
*/
DWORD NativeDataBufferWrap::GetLength()
{
	DWORD len = 0;
	HRESULT hr = pNSBuffer->GetLength(&len);
	if(FAILED(hr))
	{
		DBG_TEXT("GetLength failed with error %#08x", hr);
	}
	return len;
}

/** Queries the maximum size the buffer can be set to by SetLength

	\returns Total number of bytes the buffer spans
*/
DWORD NativeDataBufferWrap::GetMaxLength()
{
	DWORD len = 0;
	HRESULT hr = pNSBuffer->GetMaxLength(&len);
	if(FAILED(hr))
	{
		DBG_TEXT("GetLength failed with error %#08x", hr);
	}
	return len;
}

/** Updates the number of bytes used in this buffer.
	Used by the Windows Media code to know how many bytes to copy
	from the buffer

	\param newLength Bytes used
*/
void NativeDataBufferWrap::SetLength(DWORD newLength)
{
	HRESULT hr = pNSBuffer->SetLength(newLength);
	if(FAILED(hr))
	{
		DBG_TEXT("SetLength failed with error %#08x", hr);
	}
}

/** Allocates a NativeDataBufferWrap object around a native pointer
	
	\param pBuffer The buffer interface to wrap
	\returns The interface pointer to the new data buffer
*/
AudRecordCLI::IDataBuffer^ WrapNativeBuffer(INSSBuffer* pBuffer)
{
	return gcnew NativeDataBufferWrap(pBuffer);
}

using AudRecordCLI::IDataBuffer;

/** Constructs a native wrapper around the managed IDataBuffer

	\param pBuffer The managed data buffer to track via a GCHandle
*/
ManagedDataBufferWrap::ManagedDataBufferWrap(IDataBuffer^ pBuffer)
	: pDataBuffer(pBuffer),
	refCount(1)
{}

/** Returns a pointer to an interface on the current object

	\param iid Identifier of the interface to return
	\param [out] ppv Pointer to a pointer that receives the interface, or NULL
	\returns S_OK on success, E_POINTER or E_NOINtERFACE on failure
*/
STDMETHODIMP ManagedDataBufferWrap::QueryInterface(REFIID iid, PVOID* ppv)
{
	if(!ppv) return E_POINTER;
	if(iid == IID_INSSBuffer || iid == IID_IUnknown)
	{
		AddRef();
		*ppv = this;
		return S_OK;
	}
	*ppv = NULL;
	return E_NOINTERFACE;
}

/** Increments the objects reference count

	\returns The new reference count
*/
STDMETHODIMP_(ULONG) ManagedDataBufferWrap::AddRef()
{
	return _InterlockedIncrement(&refCount);
}

/** Drecreases the objects reference count

	If the ref count reaches zero, the object is destroyed
	\returns The new reference count
*/
STDMETHODIMP_(ULONG) ManagedDataBufferWrap::Release()
{
	LONG newCount = _InterlockedDecrement(&refCount);
	if(newCount <= 0)
	{
		delete this;
	}
	return static_cast<ULONG>(newCount);
}

/** Queries the current number of bytes used in the buffer,
	set by a previous call to SetLength

	\param [out] pLength Number of bytes used
	\returns S_OK on success, E_POINTER if pLen is null
*/
STDMETHODIMP ManagedDataBufferWrap::GetLength(DWORD* pLength)
{
	if(!pLength) return E_POINTER;
	*pLength = pDataBuffer->GetLength();
	return S_OK;
}

/** Updates the number of bytes used in this buffer.
	Simply forwards the value to the managed IDataBuffer::SetLength()

	\param newLength The new value of the number of bytes used
	\returns Always S_OK
*/
STDMETHODIMP ManagedDataBufferWrap::SetLength(DWORD newLength)
{
	pDataBuffer->SetLength(newLength);
	return S_OK;
}

/** Queries the maximum number of bytes that can be used in the buffer.

	\param [out] pMaxLen Number of bytes used that can be used
	\returns S_OK on success, E_POINTER if pLen is null
*/
STDMETHODIMP ManagedDataBufferWrap::GetMaxLength(DWORD* pMaxLen)
{
	if(!pMaxLen) return E_POINTER;
	*pMaxLen = pDataBuffer->GetMaxLength();
	return S_OK;
}

/** Gets a pointer to the managed buffer and unpacks the
	reutrned IntPtr to a BYTE pointer that is returned in the parameter

	\param [out] ppBuffer A pointer that receives the data buffer pointer
	\returns S_OK on success, E_POINTER if ppDataBuffer is NULL, or E_FAIL
	if the managed GetBuffer returns IntPtr::Zero
*/
STDMETHODIMP ManagedDataBufferWrap::GetBuffer(BYTE** ppBuffer)
{
	if(!ppBuffer) return E_POINTER;
	IntPtr pArray = pDataBuffer->GetBuffer();
	HRESULT hr = S_OK;
	if(pArray != IntPtr::Zero)
	{
		*ppBuffer = static_cast<BYTE*>(pArray.ToPointer());
	}
	else hr = E_FAIL;
	return hr;
}

/** Gets a pointer to the managed buffer and unpacks the
	reutrned IntPtr to a BYTE pointer that is returned in the parameter

	\param [out] ppDataBuffer A pointer that receives the data buffer pointer
	\param [out] pLength Pointer that holds the size of the buffer on return
	\returns S_OK on success, E_POINTER if either parameter is NULL, or E_FAIL
	if the managed GetBufferAndLength returns IntPtr::Zero
*/
STDMETHODIMP ManagedDataBufferWrap::GetBufferAndLength(BYTE** ppDataBuffer, DWORD* pLength)
{
	if(!(ppDataBuffer && pLength)) return E_POINTER;
	IntPtr pArray = pDataBuffer->GetBufferAndLength(*pLength);
	return (pArray != IntPtr::Zero) ? S_OK : E_FAIL;
}

/** Does the business of allocating a wrapper for a managed DataBuffer.
	If an exception is thrown by new, it's caught and logged before
	NULL is returned

	\param pBuffer The buffer to wrap
	\returns An INSSBuffer interface on a new ManagedDataBufferWrap object
	or NULL if allocation fails
*/
INSSBuffer* WrapManagedBuffer(AudRecordCLI::IDataBuffer^ pBuffer)
{
	INSSBuffer* pBufferWrap = NULL;
	try
	{
		pBufferWrap = new ManagedDataBufferWrap(pBuffer);
	}
	catch(const std::exception& e)
	{
		pBufferWrap = NULL;
		DBG_TEXT("Caught exception %s allocating a ManagedDataBufferWrap", e.what());
	}
	return pBufferWrap;
}

/** \endcond */
