/** \file DataBufferWrap.h
	\brief Contains declarations of the managed versions
	and wrappers of the INSSBuffer interface
*/

#ifndef AUDRECORDCLI_DATABUFFERWRAP_H
#define AUDRECORDCLI_DATABUFFERWRAP_H

#pragma once

using namespace System::Runtime::InteropServices;

namespace AudRecordCLI
{
	/** The managed version of the INSSBuffer interface that is
		used by the managed ITargetSink (Native IWMWriterSink) 
	*/
	public interface struct IDataBuffer
	{
	public:
		/** Gets a pointer to the buffer wrapped by the object

			\returns An IntPtr that wraps a memory buffer pointer
			that can be accessed by native code
		*/
		virtual IntPtr GetBuffer() = 0;
		/** Gets a pointer to the buffer wrapped by the object and
			its used length

			\param [out] length Number of bytes used in the returned buffer
			\returns An IntPtr that wraps a memory buffer pointer
			that can be accessed by native code
		*/
		virtual IntPtr GetBufferAndLength([Out] DWORD% length) = 0;
		/** Queries the length of the wrapped buffer

			\returns Number of used bytes in the buffer
		*/
		virtual DWORD GetLength() = 0;
		/** Queries the maximum length of the buffer

			\returns Number of bytes that the buffer pointer spans
		*/
		virtual DWORD GetMaxLength() = 0;
		/** Sets the used bytes 

			\param newUsedLength Number of bytes the user of the buffer used
		*/
		virtual void SetLength(DWORD newUsedLength) = 0;
	};
}

/** \cond DEV */
/**	Simple managed wrapper around a native INSSBuffer,
	when one needs to be passed to a method of a ITargetSink
*/
ref class NativeDataBufferWrap : public AudRecordCLI::IDataBuffer
{
private:
	/** Buffer interface that's been wrapped */
	INSSBuffer* pNSBuffer;
	/** The managed value returned by GetObjectForIUnknown */
	Object^ interfaceWrap;
public:
	NativeDataBufferWrap(INSSBuffer* pBuffer);
	virtual IntPtr GetBuffer();
	virtual IntPtr GetBufferAndLength([Out] DWORD% length);
	virtual DWORD GetLength();
	virtual DWORD GetMaxLength();
	virtual void SetLength(DWORD newLength);
};

/** Despite its name, this is a native object.
	It wraps an IDataBuffer when a INSSBuffer needs
	to be passed to a native function. This is usually where
	an IDataBuffer is returned from an ITargetSink method
*/
class ManagedDataBufferWrap : public INSSBuffer
{
private:
	/** Native handle to the wrapped databuffer */
	gcroot<AudRecordCLI::IDataBuffer^> pDataBuffer;
	/** reference count */
	LONG refCount;
public:
	ManagedDataBufferWrap(AudRecordCLI::IDataBuffer^ pBuffer);
	// IUnknown
	STDMETHODIMP QueryInterface(REFIID riid, PVOID* ppv);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();
	// INSSBuffer
	STDMETHODIMP GetLength(DWORD* pLength);
	STDMETHODIMP SetLength(DWORD newLength);
	STDMETHODIMP GetMaxLength(DWORD* pMaxLen);
	STDMETHODIMP GetBuffer(BYTE** ppBuffer);
	STDMETHODIMP GetBufferAndLength(BYTE** ppDataBuffer, DWORD* pLength);
};

AudRecordCLI::IDataBuffer^ WrapNativeBuffer(INSSBuffer* pBuffer);
INSSBuffer* WrapManagedBuffer(AudRecordCLI::IDataBuffer^ pBuffer);

/** \endcond */

#endif
