/** \file Error.cpp
	\brief Implements exceptions and error related functions
	\cond DEV
*/
#include "stdafx.h"
#include "error.h"

namespace AudRecordCLI
{
	/** Constructs a generic exception for unrecognized HRESULTs
		returned from the aative functions
		\param additionalText Description of what caused the error
		\param hr The HRESULT returned from the native function
	*/
	AudRecordException::AudRecordException(String^ additionalText, HRESULT hr)
		: Exception(additionalText)
	{
		HResult = hr;
	}
}

/** Maps a hresult to the closest managed exception class
	or to a AudRecordException if no mapping is found

	\param hr The hresult to map
*/
void ThrowExceptionIfRequired(HRESULT hr)
{
	using AudRecordCLI::AudRecordException;

	if(FAILED(hr))
	{
		String^ error = FindErrorString(hr);
		switch(hr)
		{
			case E_POINTER:
			{
				throw gcnew ArgumentNullException(error);
			}
			break;
			case E_OUTOFMEMORY:
			{
				throw gcnew OutOfMemoryException(error);
			}
			break;
			case E_INVALIDARG:
			{
				throw gcnew ArgumentException(error);
			}
			break;
			case E_FAIL:
			case E_UNEXPECTED:
			default:
			{
				throw gcnew AudRecordException(error, hr);
			}
			break;
		}
	}
}

/** Searches a module's message table for a string matching a certain id

	\param hMod The module whose message table is to be searched
	\param messageId The message id to find
	\param [out] errorText The message contents if the message is found
	\returns Whether the message was found or not
*/
BOOL GetModuleMessage(HMODULE hMod, DWORD messageId, String^& errorText)
{
	return GetModuleMessage(hMod, messageId, errorText, static_cast<va_list*>(NULL));
}

/** Finds a string in AudRecordCLI's message resources and creates a managed version
	of it. 

	\param stringId The identifier of the message
	\returns A managed version of the module string or nullptr if its not found
*/
String^ GetModuleString(DWORD stringId)
{
	LPCWSTR pwszString = NULL;
	int len = LoadString(g_hModule, stringId, reinterpret_cast<LPWSTR>(&pwszString), 0);
	return len ? gcnew String(pwszString, 0, len) : nullptr;
}

/** Searches the system message table for an error string
	that corresponds to a certain HRESULT. If that fails
	it searches the Windows Media error string dll wmerror.dll

	\param hr The hresult to message-ise
	\returns The message string if the hresult was found, or nullptr of not
*/
String^ FindErrorString(HRESULT hr)
{
	String^ messageText = nullptr;
	DWORD errorId = static_cast<DWORD>(hr);
	if(!GetModuleMessage<va_list>(NULL, errorId, messageText, NULL))
	{
		// NS_E_ error strings are in wmerror.dll
		WinType<HMODULE> wmErrorLib(LoadLibrary(L"wmerror.dll"), &FreeLibrary);
		GetModuleMessage<va_list>(*wmErrorLib, errorId, messageText, NULL);
	}
	return messageText;
}

/** \endcond */
