/** \file AudRecordCLI\stdafx.h
	\brief Precompiled header containing project wide definition
	macros and functions
	\cond DEV
*/

#ifndef STDAFX_H
#define STDAFX_H

#pragma once

#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0600
#define NOMINMAX

#include <windows.h>
#include "../macroutils.h"
#include "../exports.h"
#include "../scoped.h"
#include <vcclr.h>
#include "resource.h"
#include <intrin.h>
#include "../ETWEvents.h"

using namespace System;
using namespace System::Collections::Generic;

/** The address of the module, copy of the value given to DllMain */
extern HMODULE g_hModule;

/** If arg is null, throws a ArgumentNullException using
	the supplied name and error message

	\tparam T The type of the managed reference
	\param arg The parameter to test for null
	\param argName The name of the argument to the function
	\param errorTextMessage The id of the message in the dlls message resource
*/
template<class T>
void ThrowOnNullPtrArgument(T^ arg, LPCWSTR argName, DWORD errorTextMessage)
{
	if(!arg)
	{
		String^ message = GetModuleString(errorTextMessage);
		throw gcnew System::ArgumentNullException(gcnew String(argName), message);
	}
}

/** \def THROW_ON_NULL_ARG(arg, messageId)
	Calls ThrowOnNullPointerArg passing in \a arg, a stringified version 
	of arg (hence why this is a macro), and \a messageId
*/
#define THROW_ON_NULL_ARG(arg, messageId) ThrowOnNullPtrArgument(arg, L#arg, messageId)

/** \def LOG_TEXT(Function, string, ...)
	Formats \a string and the additional args in a sprintf manner
	and outputs it to the debug stream using ODPrint() and 
	to an ETW event stream using \a Function.
	String must be a narrow string literal, this is asserted by the macro
*/
#define LOG_TEXT(Function, string, ...) \
	do \
	{ \
		C_ASSERT(sizeof(*string) == sizeof(char)); \
		char buffer[300]; \
		char fnName[] = __FUNCTION__; \
		LogFormat(buffer, __FUNCTION__  " " string "\n", __VA_ARGS__); \
		/* The function name is already in the buffer for the ODPrint buffer */ \
		/* So we skip it when passing it to the 'Function' function */ \
		Function(__FILE__, __LINE__, __FUNCTION__, buffer + ARRAYSIZE(fnName) + 1); \
		OutputDebugStringA(buffer); \
	} \
	while(0)

/** \def SEVERE_TEXT(string, ...)
	Formats \a string and the additional args in a sprintf manner
	and outputs it to the ETW event and debug streams using LogSevereEvent()
	and ODPrint(). String must be a narrow string literal, 
	this is asserted by the macro
*/
#define SEVERE_TEXT(string, ...) \
	LOG_TEXT(LogSevereEvent, string, __VA_ARGS__)

/** \def DBG_TEXT(string, ...)
	Equivalent to the SEVERE_TEXT macro except that it logs to the 
	debug channel instead of the severe one.
*/
#define DBG_TEXT(string, ...) \
	LOG_TEXT(LogDebugEvent, string, __VA_ARGS__)

/** \def MANAGED_PRAGMA(type)
	Pushes and sets the managed pragma to \a type
	Should be followed by the MANAGED_PRAGMA_RESTORE()
	which will restore the original managed pragma setting
*/
#define MANAGED_PRAGMA(type) \
	__pragma(managed(push)) \
	__pragma(managed(type))

/** \def MANAGED_PRAGMA_RESTORE()
	Pops the managed pragma setting set by a previous 
	incantation of MANAGED_PRAGMA()
*/
#define MANAGED_PRAGMA_RESTORE() \
	__pragma(managed(pop))

/** \endcond */

#endif
