/** \file AudRecordInt.cpp	
	\brief Implements the IAudRecord interface in an AudRecord object, dev documentation
	\cond DEV
*/

#include "stdafx.h"
#include "AudRecordInt.h"
#include <mmsystem.h>
#include <mmdeviceapi.h>
#include <audioclient.h>
#include <avrt.h>
#include <algorithm>
#include "wmfunctions.h"
#include "exports.h"
#include "messages.h"
#include "misc.h"
#include "nullsink.h"

// tags for the file/network/custom sink cookies
static const DWORD FILE_SINK_TAG = 1;
static const DWORD NETWORK_SINK_TAG = 2;
static const DWORD CUSTOM_SINK_TAG = FILE_SINK_TAG | NETWORK_SINK_TAG;
static const DWORD VALID_SINK_TAGS = CUSTOM_SINK_TAG;

static PVOID ObjectToCookie(PVOID pObj, DWORD tag);
static PVOID CookieToObject(PVOID pCookie, DWORD& tag);

/** Default name used as the author of the recording stream */
static const WCHAR g_wszAuthorString[] = L"AudRecordLib";
/** the default value of the URL property of the ASF stream */
static const WCHAR g_wszAuthorURLString[] = L"http://www.airesoft.co.uk/";

/** Parameters passed as the args to the capture thread
*/
struct CreateThreadParams
{
	/** An instance pointer, used to call CaptureThreadProc() */
	AudRecord* pThis;
	/** Event to be set when all setup is complete */
	HANDLE hStartedEvent;
};

/** Initializes an AudRecord object

	Creates the exit and samples waiting events. Takes a reference to
	IWMWriter and IWMWriterAdvanced interfaces used to write the compressed 
	samples. Sets the URL and author attributes of any created output streams
	to the airesoft website and AudRecordLib respectively

	\param pWriter WMA writer object to use
	\param pWriterAdv Advanced interface queried from pWriter
*/
AudRecord::AudRecord(IWMWriter* pWriter, IWMWriterAdvanced* pWriterAdv)
	: formats(),
	hCaptureThreadExitEvent(CreateEvent(NULL, FALSE, FALSE, NULL), &CloseHandle),
	hCaptureThreadSamplesWaiting(CreateEvent(NULL, FALSE, FALSE, NULL), &CloseHandle),
	pWriter(pWriter),
	pWriterAdv(pWriterAdv),
	pAudioClient(NULL),
	pCaptureClient(NULL),
	refCount(1),
	captureThreadId(0),
	samplesWaitingTimerId(0)
{
	pWriter->AddRef();
	pWriterAdv->AddRef();
	pWriterAdv->SetLiveSource(TRUE);
}

/** Cleans up the objects resources

	Stops any recording currently active and releases all resources
*/
AudRecord::~AudRecord()
{
	ExclusiveRtlResourceLock lock(memberLock);
	Stop();
}

/** Returns pointers for implemented interfaces on the object

	Currently supported interfaces are IUnknown and IAudRecord
	\param iid The id of the requested interface
	\param [out] ppObj A pointer to a pointer that receives the interface pointer or NULL
	\returns S_OK on success, E_NOINTERFACE for unsupported interfaces and E_POINTER if ppObj is NULL
*/
STDMETHODIMP AudRecord::QueryInterface(REFIID iid, PVOID* ppObj)
{
	if(!ppObj)
	{
		return E_POINTER;
	}
	*ppObj = NULL;
	if((iid == IID_IUnknown) || (iid == IID_IAudRecord))
	{
		*ppObj = this;
	}
	if(*ppObj)
	{
		AddRef();
		return S_OK;
	}
	return E_NOINTERFACE;
}

/** Increments the objects reference count

	\returns The new reference count
*/
STDMETHODIMP_(ULONG) AudRecord::AddRef()
{
	return _InterlockedIncrement(&refCount);
}

/** Decreases the objects reference count

	If the ref count reaches zero, the object is destroyed
	\returns The new reference count
*/
STDMETHODIMP_(ULONG) AudRecord::Release()
{
	LONG newCount = _InterlockedDecrement(&refCount);
	if(newCount <= 0)
	{
		delete this;
	}
	return static_cast<ULONG>(newCount);
}

/** Checks whether the writer object can write the input samples

	DetermineAudioFormats tries to find an agreement between the mixer's sample format
	and the writer's codec input format. Despite this returning success, disagreements
	are sometimes not found by the writer until BeginWriting is called. This method
	temporarily removes all existant target sinks, adds in a null one, 
	and calls BeginWriting() to see if it would succeed. The original target sinks are
	replaced and the code from BeginWriting returned.

	\returns S_OK on success, an NS_E_* error code from the writer otherwise 
*/
HRESULT AudRecord::CheckWriterCanWrite()
{
	HRESULT hr = S_OK;
	// first, get pointers to all sinks and remove them from the object
	DWORD numSinks = 0;
	ExclusiveRtlResourceLock lock(memberLock);
	hr = pWriterAdv->GetSinkCount(&numSinks);
	// if there are no sinks set up, fail
	if(FAILED(hr) || (numSinks == 0))
	{
		return E_UNEXPECTED;
	}
	std::vector<ComType<IWMWriterSink> > sinks(numSinks);
	for(DWORD i = 0; i < numSinks; ++i)
	{
		RETURN_ON_FAILED_EXP(pWriterAdv->GetSink(i, &sinks[i]));
	}
	// removes all sinks
	RETURN_ON_FAILED_EXP(pWriterAdv->RemoveSink(NULL));

	// now all the added sinks have been removed
	// add out null sink and test if the writer can write
	NullSink nullSink(TRUE);
	// can't return if this fails, we need to replace the sinks
	if(SUCCEEDED(pWriterAdv->AddSink(&nullSink)))
	{
		hr = pWriter->BeginWriting();
		pWriter->EndWriting();
	}
	else hr = E_FAIL;
	// now we've got what we're after, remove the null sink
	// and add back the original sinks
	pWriterAdv->RemoveSink(NULL);
	for(DWORD i = 0; i < numSinks; ++i)
	{
		pWriterAdv->AddSink(*(sinks[i]));
	}
	return hr;
}

/** Sets the recording process in motion

	Creates the capture thread, sets the title of the recording stream to the current date and time,
	and checks the writer can actually write the sample format before starting the recording loop.

	If pEndPoint is not NULL, the function also sends a message to the capture thread to create
	the IAudioClient and IAudioCaptureClient interfaces and uses the returned mixer format to
	determine agreeable device and codec formats

	\params pEndpoint Optional input specifying the audio device to capture from.
	if this is NULL, it is assumed that an audio device has already been set
	in pAudioClient
	\returns S_OK on success, a HRESULT_FROM_WIN32 error if the thread cannot be created, E_FAIL 
	if the recording process has already been started.
	\sa SendThreadMessage(), DetermineAudioFormats(), CheckWriterCanWrite(), SetWriterAttributes()
*/
HRESULT AudRecord::BeginRecording(IMMDevice* pEndpoint)
{
	// if we're not being passed an endpoint, we should already have an audio client
	// and vice versa
	ASSERT((pEndpoint && (!pAudioClient)) || ((!pEndpoint) && pAudioClient));
	DWORD threadId;
	WinType<> hReadyEvent(CreateEvent(NULL, TRUE, FALSE, NULL), &CloseHandle);
	CreateThreadParams createParams = {this, *hReadyEvent};
	HANDLE hThreadTemp = CreateThread(NULL, 0, &CaptureThreadProcTrampoline, &createParams, CREATE_SUSPENDED, &threadId);
	if(!hThreadTemp)
	{
		DWORD err = GetLastError();
		return HRESULT_FROM_WIN32(err);
	}
	WinType<> hThread(hThreadTemp, &CloseHandle);
	std::wstring timeDateString = GetCurrentDateTime();
	Attribute attrs[] = {
		{g_wszWMAuthor, WMT_TYPE_STRING, g_wszAuthorString, sizeof(g_wszAuthorString)},
		{g_wszWMAuthorURL, WMT_TYPE_STRING, g_wszAuthorURLString, sizeof(g_wszAuthorURLString)},
		{g_wszWMTitle, WMT_TYPE_STRING, timeDateString.c_str(), static_cast<WORD>(timeDateString.size() * sizeof(WCHAR))}
	};

	HRESULT hr = S_OK;
	CreateAudioClientParams params = {AUDIO_BUFFER_DURATION, pEndpoint};
	ExclusiveRtlResourceLock lock(memberLock);
	// if we're being called from Record(), it's possible for more than one
	// thread to pass the if(pAudioClient) test in it.
	// Since we're serialized here, we check that somebody hasn't beaten us
	// and initialized things while we were waiting, and that we can resume the thread
	if((pEndpoint && pAudioClient) || (ResumeThread(*hThread) == MAXDWORD))
	{
		TerminateThread(*hThread, 0);
		return E_FAIL;
	}
	// wait for the thread to exit (on failure)
	// or for setup to be complete
	HANDLE hObjects[] = {*hReadyEvent, *hThread};
	DWORD waitRet = WaitForMultipleObjects(ARRAYSIZE(hObjects), hObjects, FALSE, INFINITE);
	if(waitRet == (WAIT_OBJECT_0 + 1))
	{
		// thread setup failed
		DWORD err;
		GetExitCodeThread(*hThread, &err);
		hr = HRESULT_FROM_WIN32(err);
		goto cleanup;
	}
	// create the new AudioClient if required
	if(pEndpoint)
	{
		// create it on the capure thread
		if(!(SendThreadMessage(threadId, MSG_CREATE_AUDIO_CLIENT, &params) && SUCCEEDED(params.hr)))
		{
			DWORD err = GetLastError();
			hr = params.hr ? params.hr : HRESULT_FROM_WIN32(err);
			goto cleanup;
		}
		hr = DetermineAudioFormats(params.pwfx);
	}
	// sommetimes everything goes fine until you try and start writing
	// at which point it tells you that it can't write the output format
	// or take the specific input format, despite returning success
	// from SetInputProps in DetermineAudioFormats. We check that here
	if(SUCCEEDED(hr))
	{
		hr = CheckWriterCanWrite();
	}
	// no longer need the mixer wave format
	if(params.pwfx)
	{
		CoTaskMemFree(params.pwfx);
		params.pwfx = NULL;
	}
	if(FAILED(hr))
	{
		goto cleanup;
	}
	// these have to be set before writing begins
	SetWriterAttributes(attrs, ARRAYSIZE(attrs));
	// finally start recording
	if(FAILED(hr = pWriter->BeginWriting()))
	{
		SEVERE_TEXT("CheckWriterCanWrite succeeded but BeginWriting failed with error %#08x", hr);
		goto cleanup;
	}
	// everything successful, assign data to the members
	if(params.pClient)
	{
		pAudioClient = params.pClient;
		pCaptureClient = params.pCaptureClient;
	}
	hCaptureThread.Swap(hThread);
	captureThreadId = threadId;
	if(FAILED(hr = StartRecordingLoop()))
	{
		;
	}
	return hr;

cleanup:
	if(params.pClient)
	{
		ReleaseAudioClientParams releaseParams = {params.pClient, params.pCaptureClient};
		SendThreadMessage(threadId, MSG_RELEASE_AUDIO_CLIENT, &releaseParams);
	}
	if(params.pwfx)
	{
		CoTaskMemFree(params.pwfx);
	}
	SignalObjectAndWait(*hCaptureThreadExitEvent, *hThread, INFINITE, FALSE);
	return hr;
}

/** Public entrypoint to BeginRecording

	Checks whether recording is already in progress before calling BeginRecording()
	\param pDeviceEndpoint The endpoint to begin recording from
	\returns S_OK on success, E_POINTER if pDeviceEndpoint is NULL, E_UNEXPECTED if recording is already in progress
	or a value returned by BeginRecording()
*/
STDMETHODIMP AudRecord::Record(Endpoint* pDeviceEndpoint)
{
	if(!pDeviceEndpoint)
	{
		return E_POINTER;
	}
	{
		SharedRtlResourceLock lock(memberLock);
		// if we're already initialized. fail
		if(pAudioClient)
		{
			return E_UNEXPECTED;
		}
	}
	DBG_TEXT("Recording from device %S", pDeviceEndpoint->pName);
	return BeginRecording(pDeviceEndpoint->pEndPoint);
}

/** Changes the stream composition of the writer

	Creates a new Windows Media profile, adds the specified codec stream configuration to it
	and sets it as the new profile for the writer to use. If recording is currently in progress
	it is stopped, before being restarted after the profile is changed.

	\param pNewCodec A codec configuration structure returned by GetWMACodecConfigs()
	that describes the new compression format
	\returns S_OK on success, E_POINTER if pNewCodec or pNewCodec->pConfig is NULL, 
	or a NS_E_* error from the WMA functions and interfaces.
*/
STDMETHODIMP AudRecord::ChangeCodec(CodecConfiguration* pNewCodec)
{
	if(!(pNewCodec && pNewCodec->pConfig))
	{
		return E_POINTER;
	}
	DBG_TEXT("Changing codec to %S", pNewCodec->pName);
	ComType<IWMProfileManager> pManager;
	RETURN_ON_FAILED_EXP(WMCreateProfileManager(&pManager));

	ComType<IWMProfile> pProfile;
	RETURN_ON_FAILED_EXP(pManager->CreateEmptyProfile(WMT_VER_9_0, &pProfile));
	pProfile->SetName(L"Audio");

	IWMStreamConfig* pConfig = pNewCodec->pConfig;
	pConfig->SetStreamName(L"Audio1");
	pConfig->SetStreamNumber(1);
	RETURN_ON_FAILED_EXP(pProfile->AddStream(pConfig));

	HRESULT hr;
	{
		ExclusiveRtlResourceLock lock(memberLock);
		// if we're currently recording, stop, change profile and restart
		if(pAudioClient)
		{
			// first stop recording
			StopRecordingLoop();
			pWriter->EndWriting();

			// change profile
			hr = pWriter->SetProfile(*pProfile);
			// then restart, regardless of how the set profile faired
			pWriter->BeginWriting();
			StartRecordingLoop();
		}
		// otherwise, if we're not recording, just change the profile
		else
		{
			hr = pWriter->SetProfile(*pProfile);
		}
	}
	DBG_TEXT("Writer->SetProfile returned with sttaus %#08x", hr);
	return hr;
}

/** Returns the current pause state of the recording session

	\param [out] pCurState Pointer to a BOOL that receives a nonzero value if the stream is recording
	or zero if the stream is paused
	\returns S_OK on success, E_UNEXPECTED if recording isn't currently active and E_POINTER if pCurState is NULL
*/
STDMETHODIMP AudRecord::QueryPause(AudPausedState *pCurState)
{
	if(!pCurState)
	{
		return E_POINTER;
	}
	HRESULT hr = S_OK;
	MMRESULT timerId = 0;
	{
		SharedRtlResourceLock lock(memberLock);
		// see if we're initialized
		if(!pAudioClient)
		{
			hr = E_UNEXPECTED;
		}
		// see if we're currently recording
		else 
		{
			timerId = samplesWaitingTimerId;
		}
	}
	if(SUCCEEDED(hr))
	{
		*pCurState = (timerId == 0) ? AudPaused : AudRecording;
	}
	return hr;
}

/** Pauses or resumes the recording stream

	\param [out] pCurState Optional pointer to a BOOL that receives a nonzero value if the stream is recording
	or zero if the stream is paused as a result of this function
	\returns S_OK on success or E_UNEXPECTED if recording isn't currently active
*/
STDMETHODIMP AudRecord::TogglePause(AudPausedState* pCurState)
{
	AudPausedState wasPaused;
	HRESULT hr;
	if(SUCCEEDED(hr = QueryPause(&wasPaused)))
	{
		if(wasPaused)
		{
			hr = StartRecordingLoop();
		}
		else
		{
			StopRecordingLoop();
		}
		if(SUCCEEDED(hr) && pCurState)
		{
			*pCurState = (wasPaused == AudRecording) ? AudPaused : AudRecording;
		}
	}
	return hr;
}

/** Stops the recording process

	Stops the recording loop, stops all writing to the targets,
	releases the audio and capture client interfaces and ends the capture thread
	\returns S_OK
*/
STDMETHODIMP AudRecord::Stop()
{
	ExclusiveRtlResourceLock lock(memberLock);
	if(!(pAudioClient && hCaptureThread))
	{
		return S_OK;
	}
	StopRecordingLoop();
	// ensure the handle is released
	// and the member holds nothing
	WinType<> emptyThread;
	emptyThread.Swap(hCaptureThread);

	ReleaseAudioClientParams params = {pAudioClient, pCaptureClient};
	SendThreadMessage(captureThreadId, MSG_RELEASE_AUDIO_CLIENT, &params);
	pAudioClient = NULL;
	pCaptureClient = NULL;
	captureThreadId = 0;

	SignalObjectAndWait(*hCaptureThreadExitEvent, *emptyThread, INFINITE, FALSE);

	return S_OK;
}

/** Adds a file output to the recording process

	Creates a new Windows Media file sink and points it to pwszFileName.
	Then under the member lock, adds it as a recording target
	\param pwszFileName A file name, probably can be anything openable by CreateFile
	like pipes but I haven't tested anything but normal files
	\param [out] ppCookie Pointer that receives the result of ObjectToCookie()
	for the file sink pointer
	\returns See IAudRecord::AddFileTarget()
*/
STDMETHODIMP AudRecord::AddFileTarget(LPCWSTR pwszFileName, PVOID* ppCookie)
{
	if(!(pwszFileName && ppCookie))
	{
		return E_POINTER;
	}
	// this doesn't need releasing on success as 
	// the writer maintains it's own reference
	ComType<IWMWriterFileSink> pFileSink;
	RETURN_ON_FAILED_EXP(WMCreateWriterFileSink(&pFileSink));
	RETURN_ON_FAILED_EXP(pFileSink->Open(pwszFileName));
	PVOID pCookie = ObjectToCookie(*pFileSink, FILE_SINK_TAG);
	HRESULT hr;
	{
		ExclusiveRtlResourceLock lock(memberLock);
		hr = pWriterAdv->AddSink(*pFileSink);
	}
	if(SUCCEEDED(hr))
	{
		*ppCookie = pCookie;
	}
	return hr;
}

/** Adds a network port to the recording process

	Creates a new Windows Media network sink on the *pPortNumber port.
	Optionally gets the URL of connection for the caller, if this fails 
	it is not considered an error.
	Then under the member lock, adds the sink as a recording target.
	\param pPortNumber A pointer to the requested port, 0 or NULL lets the sink choose.
	\param maxClients Number of clients who can connect, 5 is the default if 0 is given
	\param [out] ppwszUrl Optional pointer to the connection URL retreived from the sink
	\param [out] ppCookie Pointer that receives the result of ObjectToCookie()
	for the network sink pointer
	\returns See IAudRecord::AddFileTarget()
*/
STDMETHODIMP AudRecord::AddNetworkBroadcastTarget(DWORD* pPortNumber, DWORD maxClients, LPWSTR* ppwszUrl, PVOID* ppCookie)
{
	if(!ppCookie)
	{
		return E_POINTER;
	}
	// pPortNumber is optional, on input it's the desired port number
	// on return it's the actual port number
	DWORD port = 0;
	DWORD* pPortToUse = (pPortNumber ? pPortNumber : &port);

	ComType<IWMWriterNetworkSink> pNetSink;
	RETURN_ON_FAILED_EXP(WMCreateWriterNetworkSink(&pNetSink));
	pNetSink->SetMaximumClients(maxClients ? maxClients : 5);
	RETURN_ON_FAILED_EXP(pNetSink->Open(pPortToUse));
	// get the full url, if the caller wanted it
	if(ppwszUrl)
	{
		DWORD urlSize = 0;
		pNetSink->GetHostURL(NULL, &urlSize);
		LPWSTR pwszUrl = static_cast<LPWSTR>(CoTaskMemAlloc(urlSize * sizeof(WCHAR)));
		if(pwszUrl)
		{
			if(SUCCEEDED(pNetSink->GetHostURL(pwszUrl, &urlSize)))
			{
				*ppwszUrl = pwszUrl;
			}
			else
			{
				CoTaskMemFree(pwszUrl);
				*ppwszUrl = NULL;
			}
		}
	}
	PVOID pCookie = ObjectToCookie(*pNetSink, NETWORK_SINK_TAG);

	HRESULT hr;
	{
		ExclusiveRtlResourceLock lock(memberLock);
		hr = pWriterAdv->AddSink(*pNetSink);
	}
	if(SUCCEEDED(hr))
	{
		*ppCookie = pCookie;
	}
	return hr;
}

/** Adds a raw target to the recording process

	Creates a cookie for the interface, takes the lock and 
	adds pTarget as a recording target
	\param pTarget A raw IWMWriterSink interface pointer
	\param [out] ppCookie Pointer that receives the result of ObjectToCookie()
	for the pTarget pointer
	\returns See IAudRecord::AddCustomTarget
*/
STDMETHODIMP AudRecord::AddCustomTarget(IWMWriterSink* pTarget, PVOID* ppCookie)
{
	if(!(pTarget && ppCookie))
	{
		return E_POINTER;
	}
	PVOID pCookie = ObjectToCookie(pTarget, CUSTOM_SINK_TAG);
	HRESULT hr;
	{
		ExclusiveRtlResourceLock lock(memberLock);
		hr = pWriterAdv->AddSink(pTarget);
	}
	if(SUCCEEDED(hr))
	{
		*ppCookie = pCookie;
	}
	return hr;
}

/** Removes the target referred to by pCookie from the recording process

	Translates the cookie back into an object value, AddRef's the interface
	so it won't be destroyed after removing it as a Writer target.
	The object is tested to see if it is a Windows Media Network sink, if so
	the port it opened is closed.
	The object is then released and destroyed
	\param pCookie The cookie for the target to remove
	\returns See IAudRecord::RemoveTarget
*/
STDMETHODIMP AudRecord::RemoveTarget(PVOID pCookie)
{
	if(!pCookie)
	{
		return E_POINTER;
	}
	DWORD tag = 0;
	PVOID pObject = CookieToObject(pCookie, tag);
	bool removed = false;
	if(pObject)
	{
		IWMWriterSink* pSink = static_cast<IWMWriterSink*>(pObject);
		// this is so the sink outlives being removed from the writer
		// this is required because we have to call Close on the network sinks
		pSink->AddRef();
		HRESULT hr = S_OK;
		{
			ExclusiveRtlResourceLock lock(memberLock);
			hr = pWriterAdv->RemoveSink(pSink);
		}
		removed = SUCCEEDED(hr);
		if(!removed)
		{
			// this doesn't happen in an ideal world
			ASSERT(FALSE);
			SEVERE_TEXT("Caller at %p tried to remove a sink and failed with error %#x", _ReturnAddress(), hr);
		}
		else if(tag == NETWORK_SINK_TAG)
		{
			ComType<IWMWriterNetworkSink> pNetSink;
			hr = pSink->QueryInterface(&pNetSink);
			ASSERT(SUCCEEDED(hr));
			pNetSink->Close();
		}
		// we addref-ed so we have to release regardless
		pSink->Release();
	}
	return removed ? S_OK : E_INVALIDARG;
}

/** Simple static wrapper used to launch the capture thread

	Simply calls CaptureThreadProc() using the argument as the this pointer
	\param pv A CreateThreadParams struct containing the this pointer of the thread creator
	and an event to set when setup is complete
	\returns See AudRecord::CaptureThreadProc()
*/
DWORD WINAPI AudRecord::CaptureThreadProcTrampoline(PVOID pv)
{
	CreateThreadParams* pParams = static_cast<CreateThreadParams*>(pv);
	AudRecord* pThis = pParams->pThis;
	return pThis->CaptureThreadProc(pParams->hStartedEvent);
}

/** Starts the timer that periodically signals the SamplesWaiting event

	Calls timeSetEvent to signal the event that the capture thread uses to get samples
	every (audioBufferDuration / 2) milliseconds. Then it calls IAudioClient->Start to
	begin Windows' collection of audio samples
	\returns S_OK on success, HRESULT_FROM_WIN32(error) an error from IAudioClient::Start on failure
*/
HRESULT AudRecord::StartRecordingLoop()
{
	DWORD currentTimerResolution = GetCurrentTimerResolution();
	MMRESULT id = 0;
	HRESULT hr = S_OK;
	{
		ExclusiveRtlResourceLock lock(memberLock);
		ASSERT(pAudioClient);
		// if the event is already set, exit
		if(samplesWaitingTimerId)
		{
			return S_OK;
		}
		id = timeSetEvent(
			HALF_BUFFER_DURATION_IN_MILLISEC,
			currentTimerResolution,
			reinterpret_cast<LPTIMECALLBACK>(*hCaptureThreadSamplesWaiting),
			0,
			TIME_PERIODIC | TIME_CALLBACK_EVENT_SET | TIME_KILL_SYNCHRONOUS
		);
		if(id != NULL)
		{
			samplesWaitingTimerId = id;
			hr = pAudioClient->Start();
		}
		else
		{
			DWORD err = GetLastError();
			hr = HRESULT_FROM_WIN32(err);
		}
	}
	DBG_TEXT("Started recording loop with timer id = %lu", id);
	return hr;
}

/** Stops the timer that continually sets the SamplesWaiting event

	Calls timeKillEvent to stop signalling the samplesWaiting event,
	then calls IAudioClient::Stop() to stop Windows buffering audio samples
	This effectively pauses recording.
*/
void AudRecord::StopRecordingLoop()
{
	MMRESULT timerId = 0;
	ExclusiveRtlResourceLock lock(memberLock);
	ASSERT(pAudioClient);
	timerId = samplesWaitingTimerId;
	samplesWaitingTimerId = 0;
	if(timerId)
	{
		timeKillEvent(timerId);
		pAudioClient->Stop();
	}
}

/** Queries for outstanding samples and converts them to the codec format

	Captures source samples for as long as there are some available from the capture client.
	The source samples are then converted to the codec input format using the function
	determined by DetermineAudioFormats(). These converteed samples are then written
	to the codec with a timecode of (framesSeen * (REFTIMES_PER_SECOND / codecInput.samplesPerSec))

	\param [in, out] framesSeen On input, this is the number of audio frames processed so far.
	It is updated during the function to provide a continuous count
	\param timePerSample The number of 100-nanosecond units in a second (REFTIMES_PER_SEC)
	divided by the number of samples per second for the codec. It is calculated is the calling
	function because it's a	constant once the sample format is known
*/
void AudRecord::CaptureSamples(UINT64& framesSeen, const double& timePerSample)
{
	UINT packetLen = 0;
	UINT failures = 0;
	HRESULT hr;
	SharedRtlResourceLock lock(memberLock);
	UINT codecInputBytes = formats.codecInput.nBlockAlign;
	// keep looping until there are unprocessed samples
	RETURN_ON_FAILED_EXP_VOID(pCaptureClient->GetNextPacketSize(&packetLen));
	while(packetLen)
	{
		ComType<INSSBuffer> pWMBuffer;
		// try and allocate a codec sample buffer, if it fails three times
		// give it up
		UINT maxDataSize = packetLen * codecInputBytes;
		if(FAILED(hr = pWriter->AllocateSample(maxDataSize, &pWMBuffer)))
		{
			DBG_TEXT("Failed to allocate writer simple, error %8x", hr);
			if((failures++) > 3)
			{
				return;
			}
			continue;
		}
		BYTE* pDstSamples;
		RETURN_ON_FAILED_EXP_VOID(pWMBuffer->GetBuffer(&pDstSamples));

		BYTE* pSrcSamples;
		DWORD flags;
		UINT numFramesAvailable;
        RETURN_ON_FAILED_EXP_VOID(pCaptureClient->GetBuffer(
           &pSrcSamples,
           &numFramesAvailable,
		   &flags,
		   NULL,
		   NULL
		));
		UINT neededDataSize = numFramesAvailable * codecInputBytes;
		double frameTime = framesSeen * timePerSample;
		if(flags & AUDCLNT_BUFFERFLAGS_SILENT)
		{
			// Luckily all bytes zero is silence for both integral
			// and floating codec inputs format
			memset(pDstSamples, 0, neededDataSize);
		}
		else
		{
			neededDataSize = sampleConverterFn(pSrcSamples, numFramesAvailable, pDstSamples);
		}
		pWMBuffer->SetLength(neededDataSize);
		if(FAILED(hr = pWriter->WriteSample(0, ULONGLONG(frameTime), 0, *pWMBuffer)))
		{
			SEVERE_TEXT("Failed to write sample, error %#08x", hr);
		}
		else
		{
			framesSeen += numFramesAvailable;
		}
		pCaptureClient->ReleaseBuffer(numFramesAvailable);
		if(FAILED(hr = pCaptureClient->GetNextPacketSize(&packetLen)))
		{
			SEVERE_TEXT("Failed to get next packet length, error %#08x", hr);
			packetLen = 0;
		}
	}
}

/** The entrypoint to the capture thread.

	Opts in to the MMCSS for optimum thread scheduling and priority for audio capture
	then waits on three things in a loop. The exit event, the pseudo windows message event
	and the APC delivery that signals samples have been collected. 
	The exit event exits the loop and thread, 
	the samples ready event causes a call to CaptureSamples(), and the 
	windows message event processes all outstanding Windows messages posted by another
	thread

	\param hHasStarted An event that is set when the thread has finished initialising
	\returns Always Zero
*/
DWORD AudRecord::CaptureThreadProc(HANDLE hHasStarted)
{
	HANDLE hWaitables[] = {*hCaptureThreadExitEvent, *hCaptureThreadSamplesWaiting};
	// Multithreaded com initialisation affects all threads 
	// created after the first thread that uses it
	// since we can be used in .Net apps, we don't use multithreaded
	// unless somebody else already has 
	if(FAILED(CoInitializeEx(NULL, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE)))
	{
		CoInitializeEx(NULL, COINIT_MULTITHREADED | COINIT_DISABLE_OLE1DDE);
	}
	DWORD taskIndex = 0;
	// opt in to the MultiMediaClassSchedulingService for fancy thread priority stuff
	// we don't use task index but it can't be NULL
	HANDLE hThreadChars = AvSetMmThreadCharacteristics(L"Capture", &taskIndex);
	DWORD ret = 0;
	BOOL exit = FALSE;
	BOOL firstTime = TRUE;
	UINT64 framesSeen = 0;
	// we can't initialize this here, as the sample formats won't have been decided yet
	double timePerSample;
	// for debugger help
	SetCurrentThreadName("AudioSampleCaptureThread");
	// ensure a message queue is created
	// if this doesn't work, there's nothing we can do but exit
	if(!IsGUIThread(TRUE))
	{
		return GetLastError();
	}
	// signal that setup is created
	SetEvent(hHasStarted);
	do
	{
		DWORD waitRet = MsgWaitForMultipleObjectsEx(
			2,
			hWaitables,
			INFINITE,
			QS_ALLEVENTS,
			MWMO_INPUTAVAILABLE
		);
		switch(waitRet)
		{
			// exit event
			case WAIT_OBJECT_0:
			{
				exit = TRUE;
			}
			break;
			// samples waiting
			case WAIT_OBJECT_0 + 1:
			{
				if(firstTime)
				{
					SharedRtlResourceLock lock(memberLock);
					timePerSample = REFTIMES_PER_SEC / static_cast<double>(formats.mixer.nSamplesPerSec);
					firstTime = FALSE;
				}
				CaptureSamples(framesSeen, timePerSample);
			}
			break;
			// message
			case WAIT_OBJECT_0 + 2:
			{
				MSG msg;
				while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
				{
					ActOnMessage(msg);
				}
			}
			break;
			default:
			{
				SEVERE_TEXT("MessageWaitForMultipleObjectsEx unexpectedly returned %lu", waitRet);
			}
			break;
		}
	}
	while(!exit);

	pWriter->EndWriting();
	AvRevertMmThreadCharacteristics(hThreadChars);
	CoUninitialize();
	return ret;
}

/** Tries to agree samples formats between the mixer and codec

	Calls MatchCodecInputFormat to determine a matching codec input to the waveformat
	with the minumum amount of information loss. Also calls DetermineSampleConverter
	to decide on a conversion function that takes mixer samples and produces
	codec samples

	\param pwfx The WAVEFORMAT on the mixer/endpoint device
	\returns S_OK on success, a value from MatchCodecInputFormat on failure
*/
HRESULT AudRecord::DetermineAudioFormats(WAVEFORMATEX *pwfx)
{
	WAVEFORMATEX& mixerFormat = formats.mixer;
	mixerFormat = *pwfx;
	BOOL isFloat = FALSE;
	if(GetWaveformatExType(pwfx, &isFloat) && isFloat)
	{
		// flag used in MatchCodecInputFormat to mean 
		// that this format has floating point samples
		mixerFormat.wFormatTag = 0;
	}
	RETURN_ON_FAILED_EXP(MatchCodecInputFormat(*pWriter, mixerFormat, formats.codecInput));
	ConverterFnPtr sampleConverterFnTemp;
	DetermineSampleConverter(mixerFormat, formats.codecInput, sampleConverterFnTemp);
	ExclusiveRtlResourceLock lock(memberLock);
	// causes ADL lookup of a more appropriate swap method 
	// if one exists than calling std::swap directly
	using std::swap;
	swap(sampleConverterFn, sampleConverterFnTemp);
	return S_OK;
}

/** Sets a Windows Media g_wszWM* attribute on the output stream

	Queries the writer for a IWMHeaderInfo3 interface and then sets 
	attributes for the stream header using the details in the Attribute
	structures. SetAttribute is used over the more recommended
	ModifyAttribute as it can both sets new and replace existing attributes.
	ModifyAttribute can only change existing ones.

	\param pAttributes An array of attributes to set
	\param numAttrs The number of entries in the pAttributes array
*/
void AudRecord::SetWriterAttributes(Attribute* pAttributes, DWORD numAttrs)
{
	ComType<IWMHeaderInfo3> pHeader;
	if(SUCCEEDED(pWriter->QueryInterface(&pHeader)))
	{
		for(DWORD i = 0; i < numAttrs; ++i)
		{
			pHeader->SetAttribute(0, pAttributes[i].pwszName, pAttributes[i].type, static_cast<const BYTE*>(pAttributes[i].pData), pAttributes[i].dataSize);
		}
	}
}

/** Gets current writing statistics from the codec

	Calls the GetStatistics method of IWMWriterAdvanced interface. If it succeeds
	the filled in WM_WRITER_STATISTICS structure is copied to *pStats as the structures
	are identical.

	\param [out] pStats A pointer that receives statistics from the writing codec
*/
HRESULT AudRecord::GetStatistics(AudRecordStats* pStats)
{
	if(!pStats)
	{
		return E_POINTER;
	}
	HRESULT hr = S_OK;
	WM_WRITER_STATISTICS stats = {0};
	{
		SharedRtlResourceLock lock(memberLock);
		hr = pWriterAdv->GetStatistics(0, &stats);
	}
	if(SUCCEEDED(hr))
	{
		memcpy(pStats, &stats, sizeof(*pStats));
	}
	return hr;
}

/** Creates an AudRecord object using the specified writer and configuration

	Queries the writer for its IWMWriterAdvanced interface then allocates a new 
	AudRecord object passing in both interfaces. ChangeCodec is then called
	on the new object to set the requested configuration.

	\param pWriter A IWMWriter interface the AudRecord object will take a reference to
	\param pCodecConfig The initial output format of the writer
	\param [out] ppRecord A pointer that recieves the IAudRecrd interface of the object
	\returns S_OK on success, E_OUTOFMEMORY if new throws an exception,
	or a value from ChangeCodec otherwise
*/
HRESULT AudRecord::Create(IWMWriter *pWriter, CodecConfiguration* pCodecConfig, IAudRecord** ppRecord)
{
	ComType<IWMWriterAdvanced> pWriterAdv;
	RETURN_ON_FAILED_EXP(pWriter->QueryInterface(&pWriterAdv));
	ComType<IAudRecord> pRecord;
	HRESULT hr = S_OK;
	try
	{
		pRecord.Init(new AudRecord(pWriter, *pWriterAdv));
		hr = pRecord->ChangeCodec(pCodecConfig);
	}
	catch(const std::bad_alloc&)
	{
		DBG_TEXT("Got out of memory exception creating AudRecord object");
		hr = E_OUTOFMEMORY;
	}
	if(SUCCEEDED(hr))
	{
		*ppRecord = pRecord.Release();
	}
	return hr;
}

/** 'Salt' added used by ObjectToCookie and CookieToObject
*/
static const DWORD g_saltVal = 0x31323330UL;

/** Converts a target sink object pointer to a cookie

	Adds a constant value and a tag to an actual IWMWriterSink object
	This is so we can pass a stable value back to the client
	so they can remove it at a later time if they wish
	it also means we don't have to hold a list of sinks ourselves

	\param pObject The object pointer to create a cookie for
	\param tag The type of input object
	\returns The cookie that corresponds to pObject
*/
PVOID ObjectToCookie(PVOID pObject, DWORD tag)
{
	// assset the object is 4 byte aligned
	// we need the 2 lowest bits to be clear for the tag
	ASSERT(!(PtrToUlong(pObject) & VALID_SINK_TAGS));
	ULONG_PTR tagValue = reinterpret_cast<ULONG_PTR>(pObject);
	tagValue += g_saltVal;
	tagValue |= tag;
	return reinterpret_cast<PVOID>(tagValue);
}

/** Converts a cookie back to the original object

	Takes a value returned from ObjectToCookie(), removes the
	'salt' and the tag and returns both the original object
	and the tag back to the caller

	\param pCookie The cookie returned by ObjectToCookie()
	\param [out] tag The tag applied to this object at cookie creation time
	\returns The original object
*/
PVOID CookieToObject(PVOID pCookie, DWORD& tag)
{
	ULONG_PTR tagValue = reinterpret_cast<ULONG_PTR>(pCookie);
	tagValue -= g_saltVal;

	tag = tagValue & VALID_SINK_TAGS;
	if(!tag)
	{
		// shouldn't happen
		ASSERT(FALSE);
		return NULL;
	}
	tagValue &= ~VALID_SINK_TAGS;
	PVOID pObject = reinterpret_cast<PVOID>(tagValue);
	return pObject;
}

/** \endcond
*/