/** \file AudRecordInt.h
	\brief Declares the AudRecord class that implements the IAudRecord interface, dev documentation
	\cond DEV
*/

#ifndef AUDRECORD_INTERFACE_H
#define AUDRECORD_INTERFACE_H

#pragma once

#include "IAudRecordInterface.h"
#include <wmsdk.h>
#include <mmreg.h>
#include <audioclient.h>
#include "rtlresource.h"
#include "convert.h"
#include "exports.h"

/** Manages the recording of audio from a specific endpoint
	and writing it to optional file, network, and custom 
	outputs in a specific codec format
*/
class AudRecord : public IAudRecord
{
	/** Holds the wave formats of the mixer output and codec input */
	struct Formats
	{
		WAVEFORMATEX mixer;
		WAVEFORMATEX codecInput;
	};

	/** Details of an attribute to set on the recording stream */
	struct Attribute
	{
		/** The name of the atribute, a g_wszWM* value */
		LPCWSTR pwszName;
		/** The type of the atribute, a WM_TYPE_* value */
		WMT_ATTR_DATATYPE type;
		/** The data to set */
		LPCVOID pData;
		/** The size of the pData buffer */
		WORD dataSize;
	};
private:
	/** Read write lock used when inspecing/modifying the members */
	RtlResource memberLock;
	/** The makeup of the samples we aer converting from and to */
	Formats formats;
	/** The function that does all the sample conversion */
	ConverterFnPtr sampleConverterFn;
	/** The thread which does the sample capturing and IAudioClient creation */
	WinType<HANDLE> hCaptureThread;
	/** The event which exits the sample thread when set */
	WinType<HANDLE> hCaptureThreadExitEvent;
	/** The event set when samples are waiting to be converted */
	WinType<HANDLE> hCaptureThreadSamplesWaiting;
	/** The base interface which accepts the converted samples 
		and writes them to the sinks/targets
	*/
	ComType<IWMWriter> pWriter;
	/** The advanced portion of the Windows Media interface which
		concerns itself with sink/target management
	*/
	ComType<IWMWriterAdvanced> pWriterAdv;
	/** Object representing out link with the audio subsystem */
	IAudioClient* pAudioClient;
	/** Links to the audio subsystem and retrieves the source samples */
	IAudioCaptureClient* pCaptureClient;
	/** References to this particular object, once it gets to 0
	the object is deleted
	*/
	LONG refCount;
	/** Numerical id of the capture thread, we need this to send messages to it */
	DWORD captureThreadId;
	/** The id if the timer that sets hCaptureThreadSamplesWaiting
		at appropriate intervals
	*/
	UINT samplesWaitingTimerId;

	AudRecord(IWMWriter* pWriter, IWMWriterAdvanced* pWriterAdv);
	~AudRecord();

	DWORD CaptureThreadProc(HANDLE hStartedThread);
	static DWORD WINAPI CaptureThreadProcTrampoline(PVOID pv);

	HRESULT DetermineAudioFormats(WAVEFORMATEX* pwfx);

	HRESULT StartRecordingLoop();
	void StopRecordingLoop();

	void CaptureSamples(UINT64& framesSeen, const double& timePerSample);

	HRESULT CheckWriterCanWrite();
	void SetWriterAttributes(Attribute* pAttributes, DWORD numAttrs);

	HRESULT BeginRecording(struct IMMDevice* pEndpoint);

	// disable copying
	AudRecord(const AudRecord&);
	void operator=(const AudRecord&);

public:
	static HRESULT Create(IWMWriter* pWriter, CodecConfiguration* pCodecConfig, IAudRecord** ppRecord);

	// IUnknown
	STDMETHODIMP QueryInterface(REFIID iid, PVOID* ppObj);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();

	// IAudRecord
	STDMETHODIMP Record(Endpoint* pClient);
	STDMETHODIMP ChangeCodec(CodecConfiguration* pNewCodec);
	STDMETHODIMP TogglePause(AudPausedState* pCurState);
	STDMETHODIMP QueryPause(AudPausedState* pCurState);
	STDMETHODIMP Stop();
	STDMETHODIMP AddFileTarget(LPCWSTR pwszFileName, PVOID* ppCookie);
	STDMETHODIMP AddNetworkBroadcastTarget(DWORD* pPort, DWORD maxConnections, LPWSTR* ppwszUrl, PVOID* ppCookie);
	STDMETHODIMP AddCustomTarget(IWMWriterSink* pTarget, PVOID* ppCookie);
	STDMETHODIMP RemoveTarget(PVOID pCookie);
	STDMETHODIMP GetStatistics(AudRecordStats* pStats);
};

#endif

/** \endcond
*/
