/** \file ETWEventFunctions.inl
	\brief Includes the common functions for ETW error reporting
	This file isn't compiled as normal, it's included by those
	who wish to use its error reporting
	\cond DEV
*/

#pragma managed(off)

#include "ETWEvents.h"

// if this fires, you probably forgot to define 
// ETW_PROVIDER_GUID as the provider guid
C_ASSERT(sizeof(ETW_PROVIDER_GUID) == sizeof(GUID));

/** \def ODS_PRINT(format, ...)
	Formats \a format and any additional args using
	LogFormat() before outputting the result using 
	OutputDebugStringA(). This is a macro
	so we can pass the variable args to LogFormat
*/
#define ODS_PRINT(format, ...) \
	{ \
		char buffer[300]; \
		LogFormat(buffer, __FUNCTION__ " " format, __VA_ARGS__); \
		OutputDebugStringA(buffer); \
	}

/** The handle to the provider, initialised in RegisterEventProvider */
static REGHANDLE g_hEventProvHandle;

/** Registers a certain provider event source with ETW

	It relies on a previously defined or existant ETW_PROVIDER_GUID	value
	that contains the GUID of the provider.
 */
void RegisterEventProvider()
{
	ULONG err = EventRegister(&ETW_PROVIDER_GUID, NULL, NULL, &g_hEventProvHandle);
	if(err != ERROR_SUCCESS)
	{
		ODS_PRINT("EventRegister failed with error %lu", err);
	}
}

/** Unregisters a certain provider event source with ETW so it can
	no longer send events
 */
void UnregisterEventProvider()
{
	ULONG err = EventUnregister(g_hEventProvHandle);
	if(err != ERROR_SUCCESS)
	{
		ODS_PRINT("EventUnregister failed with error %lu", err);
	}
	g_hEventProvHandle = NULL;
}

/** Send an event to ant listeners
	
	Descriptors are created for each parameter passed into the event, before it is written
	using the supplied description
	\param pszFile the file in which the event occurred
	\param line The line on which it occurred on
	\param pszFunction Narrow string containing the name of the function it occurred in
	\param pszdata Narrow string containing any additional text
	\param evDesc Description of the event to raise
 */
void WriteEvent(LPCSTR pszFile, DWORD line, LPCSTR pszFunction, LPCSTR pszData, const EVENT_DESCRIPTOR& evDesc)
{
	EVENT_DATA_DESCRIPTOR dataDesc[4];
	EventDataDescCreate(&dataDesc[0], pszFile, strlen(pszFile));
	EventDataDescCreate(&dataDesc[1], &line, sizeof(line));
	EventDataDescCreate(&dataDesc[2], pszFunction, strlen(pszFunction));
	EventDataDescCreate(&dataDesc[3], pszData, strlen(pszData));
	ULONG err = EventWrite(g_hEventProvHandle, &evDesc, ARRAYSIZE(dataDesc), dataDesc);
	if(err != ERROR_SUCCESS)
	{
		ODS_PRINT("EventWrite failed with error %lu", err);
	}
}

/** Creates a debug event in the log

	The function is equivalent to calling WriteEvent with FunctionFailureDebug
	as the event descriptor, but only if the event is enabled by the log
	consumer
	
	\param pszFile the file in which the event occurred
	\param line The line on which it occurred on
	\param pszFunction Narrow string containing the name of the function it occurred in
	\param pszdata Narrow string containing any additional text
*/
void LogDebugEvent(LPCSTR pszFile, DWORD line, LPCSTR pszFunction, LPCSTR pszData)
{
	if(EventEnabled(g_hEventProvHandle, &FunctionFailureDebug))
	{
		WriteEvent(pszFile, line, pszFunction, pszData, FunctionFailureDebug);
	}
}

/** Creates a severe error event in the log

	The function is equivalent to calling WriteEvent with FunctionError
	as the event descriptor, but only if the event is enabled by the log
	consumer
	
	\param pszFile the file in which the event occurred
	\param line The line on which it occurred on
	\param pszFunction Narrow string containing the name of the function it occurred in
	\param pszdata Narrow string containing any additional text
*/
void LogSevereEvent(LPCSTR pszFile, DWORD line, LPCSTR pszFunction, LPCSTR pszData)
{
	if(EventEnabled(g_hEventProvHandle, &FunctionError))
	{
		WriteEvent(pszFile, line, pszFunction, pszData, FunctionError);
	}
}

/** \endcond */
