/** \file device.cpp
	\brief Contains functions for the enumeration of audio render endpoints
	\cond DEV
*/

#include "StdAfx.h"
#include "Device.h"
#include <audioclient.h>
#include <mmdeviceapi.h>
#include <Functiondiscoverykeys_devpkey.h>
#include <algorithm>
#include <functional>

// these aren't in any of the libs
const CLSID CLSID_MMDeviceEnumerator = __uuidof(MMDeviceEnumerator);
const IID IID_IMMDeviceEnumerator = __uuidof(IMMDeviceEnumerator);
const IID IID_IAudioClient = __uuidof(IAudioClient);
const IID IID_IAudioCaptureClient = __uuidof(IAudioCaptureClient);

/** Temporary RAII holder for an endpoint and its name */
struct EndpointTemp
{
	/** Device pointer */
	ComType<IMMDevice> pDevice;
	/** Device name */
	std::wstring name;

	/** Initializes an object with a device pointer and a name

		\param pDevice The device's interface pointer to wrap
		\param name Name of the device
	*/
	EndpointTemp(IMMDevice* pDevice, const std::wstring& name)
		: pDevice(pDevice),
		name(name)
	{}

	/** Copies an EndpointTemp and takes an extra reference 
		on the device pointer

		\param other The object to copy
	*/
	EndpointTemp(const EndpointTemp& other)
		: pDevice(*other.pDevice),
		name(other.name)
	{
		pDevice->AddRef();
	}

	/** Makes this object equivalent to another

		The currently wrapped device pointer is released and made
		available for other's, which it is reinitialised with and AddRef'd.

		\param other The object to make this one equivalent to
		\returns Reference to this
	*/
	EndpointTemp& operator=(const EndpointTemp& other)
	{
		name = other.name;
		ComType<IMMDevice> pTemp(pDevice.Release());
		if(*other.pDevice)
		{
			pDevice.Init(*other.pDevice);
			pDevice->AddRef();
		}
		return *this;
	}

	/** Releases the pDevice's hold on the interface pointer
		so it's destructor won't release it
	*/
	void ReleaseInterfaceReference()
	{
		pDevice.Release();
	}
};

/** Enumerates audio output devices

	All audio devices are discovered using the multimedia device enumerator.
	For every device, a wrapped interface pointer and name are temporarily saved while a running
	count of the memory required for all output structs name strings is maintained.
	One allocation is then made and the saved pointers and names are serialised into it.
	No interface pointers in the returned struct should be Released, rather the whole thing
	should be passed to FreeAllAudioEndpoints

	\param [out] pEndpoints On return contains the enumerated audio endpoint interface pointers and names
	\returns S_OK on success, standard COM error code otherwise
*/
HRESULT GetAllAudioEndpoints(EndpointArray*& pEndpoints)
{
	ComType<IMMDeviceEnumerator> pEnum;
	RETURN_ON_FAILED_EXP(CoCreateInstance(CLSID_MMDeviceEnumerator, NULL, CLSCTX_INPROC, IID_PPV_ARGS(&pEnum)));
	ComType<IMMDeviceCollection> pColl;
	// find our output devices
	RETURN_ON_FAILED_EXP(pEnum->EnumAudioEndpoints(eAll, DEVICE_STATE_ACTIVE, &pColl));
	UINT numDevs = 0;
	RETURN_ON_FAILED_EXP(pColl->GetCount(&numDevs));

	std::vector<EndpointTemp> tempEndpoints;
	tempEndpoints.reserve(numDevs);

	DWORD requiredMemory = sizeof(EndpointArray);
	// here we go round the mulberry bush
	for(DWORD i = 0; i < numDevs; ++i)
	{
		HRESULT hr;
		// grab a pointer to this device
		ComType<IMMDevice> pDevice;
		if(FAILED(hr = pColl->Item(i, &pDevice)))
		{
			DBG_TEXT("Failed to get MMDevice, error %#08x", hr);
			continue;
		}
		// they find its real name not it's guidy windows name
		// via its property store
		ComType<IPropertyStore> pProps;
		if(FAILED(hr = pDevice->OpenPropertyStore(STGM_READ, &pProps)))
		{
			DBG_TEXT("Failed to open property store, error %#08x", hr);
			continue;
		}
		// query the name
		PROPVARIANT prop = {0};
		if(FAILED(hr = pProps->GetValue(PKEY_Device_FriendlyName, &prop)))
		{
			DBG_TEXT("Failed to get friendly name of device, error %#08x", hr);
			continue;
		}
		// then add it to the list
		std::wstring name = prop.pwszVal;
		PropVariantClear(&prop);
		tempEndpoints.push_back(EndpointTemp(pDevice.Release(), name));
		requiredMemory += (sizeof(Endpoint) + ((name.length() + 1) * sizeof(WCHAR)));
	}
	numDevs = tempEndpoints.size();
	BYTE* pMemStart = new BYTE[requiredMemory];
	EndpointArray* pArray = reinterpret_cast<EndpointArray*>(pMemStart);
	pArray->num = numDevs;

	// the first Endpoint is accounted for in the size of *pArray
	DWORD stringSpaceOffset = sizeof(*pArray) + (sizeof(Endpoint) * (numDevs - 1));

	// fill in the output structs
	WCHAR* pStringMem = reinterpret_cast<WCHAR*>(pMemStart + stringSpaceOffset);
	Endpoint* pArrayIter = pArray->outputs;
	for(DWORD i = 0; i < numDevs; ++i, ++pArrayIter)
	{
		EndpointTemp& device = tempEndpoints[i];
		// we don't release the device here, in case something later in the loop
		// causes an exception. If it does, we'll lose the pointer
		pArrayIter->pEndPoint = *device.pDevice;
		// we need to account for the null in the name
		// which isn't present in the string
		size_t endpointNameLen = device.name.length() + 1;
		StringCchCopy(pStringMem, endpointNameLen, device.name.c_str());
		pArrayIter->pName = pStringMem;
		pStringMem += endpointNameLen;
	}
	// release the wrappers hold on the interfaces
	// the ref count on all the interfaces is currently 1
	// so if we don't release our hold they'll all be freed at 
	// the end of the function
	std::for_each(tempEndpoints.begin(), tempEndpoints.end(), std::mem_fun_ref(&EndpointTemp::ReleaseInterfaceReference));
	pEndpoints = pArray;
	return S_OK;
}

/** Releases device interface pointers and frees the memory for an endpoint array

	\param pEndpoints An array of endpoints to free,
	must have been allocated by GetAllAudioEndpoints
*/
void FreeAllAudioEndpoints(EndpointArray* pEndpoints)
{
	DWORD num = pEndpoints->num;
	Endpoint* pEndpointInfo = pEndpoints->outputs;
	for(DWORD i = 0; i < num; ++i, ++pEndpointInfo)
	{
		if(pEndpointInfo->pEndPoint)
		{
			pEndpointInfo->pEndPoint->Release();
		}
		else
		{
			SEVERE_TEXT("Endpoint array has been corrupted, encountered NULL interface pointer for '%S'", pEndpointInfo->pName);
		}
	}
	delete [] reinterpret_cast<BYTE*>(pEndpoints);
}

/** \endcond */
