/** \file macroutils.h
	\brief Defines a custom ASSERT macro and other useful macro stuff
	\cond DEV
*/

#ifndef WINLIB_MACROUTILS_H
#define WINLIB_MACROUTILS_H

#pragma once

#pragma region support macros
/** \def CONCAT_INT(a, b)
	Internal macro that actually performs the concatenation of \a a and \a b 
	with the token pasting operator
*/
#define CONCAT_INT(a, b) a##b
/** \def CONCAT(a, b)
	Concatenates \a b to \a a by passing the arguments to CONCAT_INT()
	This is so any macro parameters are exapanded before the pasting
	takes place
*/
#define CONCAT(a, b) CONCAT_INT(a, b)

/** \def WSTRINGIFY_INT(a)
	Internal macro that actually turns a into a wide string
	This is achieved by passing an L and a narrow string
	of \a a to CONCAT so they are paasted together
*/
#define WSTRINGIFY_INT(a) CONCAT(L, #a)
/** \def WSTRINGIFY(a)
	Turns \a a into a wide character string by putting it
	in speech marks and prepending an L. The argument is
	passed to WSTRINGIFY_INT so that a is expanded if
	it's a macro itself
*/
#define WSTRINGIFY(a) WSTRINGIFY_INT(a)
/** \def STRINGIFY_INT(a)
	Internal macro that actually turns a into a string
	This is simply done using the stringify (#) operator
*/
#define STRINGIFY_INT(a) #a
/** \def STRINGIFY(a)
	Turns \a a into a character string by putting it
	in speech marks. The argument is
	passed to STRINGIFY_INT so that a is expanded if
	it's a macro itself
*/
#define STRINGIFY(a) STRINGIFY_INT(a)

/** \def DISABLE_CONST_COND()
	As the name suggests, the macro disables the
	'conditional expression is constant' warning
	Primarily for use in asserts and other macros
	that test conditions
*/
#define DISABLE_CONST_COND() \
	__pragma(warning(push)) \
	__pragma(warning(disable : 4127))

/** \def END_DISABLE_CONST_COND()
	Restores the compilers warning state that was 
	saved before the conditional warning was 
	disabled by DISABLE_CONST_COND()
*/
#define END_DISABLE_CONST_COND() \
	__pragma(warning(pop))

#pragma endregion

/** \def CASSERT(expr)
	Causes a compile time error by trying to define an
	array of negative size is the \a expr evaluates
	to false. 
*/
#define CASSERT(expr) \
	static const char CONCAT(cassertTester, __COUNTER__) [(expr) ? 1 : -1]

#if defined(DEBUG) || defined(_DEBUG) || defined(_DBG)

#ifdef _MSC_VER 
__if_not_exists(OutputDebugStringA)
{
	// so we don't have to include the windows header files
	extern "C" __declspec(dllimport) void __stdcall OutputDebugStringA(const char*);
	extern "C" __declspec(dllimport) void __stdcall DebugBreak();
}
#endif

/** \def ASSERT(expr)
	If \a expr is false, it logs an error to the debugger text output.
	It then calls DebugBreak to stop the program on the line that failed
	the test. In non-debug builds, it does nothing.
*/
#define ASSERT(expr) \
	DISABLE_CONST_COND() \
	do \
	{ \
		if(!(expr)) \
		{ \
			OutputDebugStringA( \
			"Assertion Failed: " \
			#expr \
			"\nFile: " \
			__FILE__ \
			", line " \
			STRINGIFY(__LINE__) \
			", function \"" \
			__FUNCTION__ \
			"\".\n" \
			); \
			DebugBreak(); \
		} \
	} \
	while(0); \
	END_DISABLE_CONST_COND()	

#else

/** \def ASSERT(expr)
	If \a expr is false, it logs an error to the debugger text output.
	It then calls DebugBreak to stop the program on the line that failed
	the test. In non-debug builds, it does nothing.
*/
#define ASSERT(expr)

#endif

/** \endcond */

#endif
