/** \file rtlresource.cpp
	\brief Implements the thin wrapper class around the fat read/write lock functions
	that are exported by ntdll.dll
	\cond DEV
*/

#include "stdafx.h"
#include "rtlresource.h"

#pragma region Function pointer bits
#pragma region Function pointer typedefs
typedef void (NTAPI*pfnRtlInitializeResource)(PRTL_RESOURCE);
typedef void (NTAPI*pfnRtlDeleteResource)(PRTL_RESOURCE);
typedef void (NTAPI*pfnRtlDumpResource)(const RTL_RESOURCE*);
typedef void (NTAPI*pfnRtlReleaseResource)(PRTL_RESOURCE);
typedef BOOLEAN (NTAPI*pfnRtlAcquireResourceExclusive)(PRTL_RESOURCE, BOOLEAN wait);
typedef BOOLEAN (NTAPI*pfnRtlAcquireResourceShared)(PRTL_RESOURCE, BOOLEAN wait);
#pragma endregion

#pragma region Function pointers
pfnRtlInitializeResource rtlInitializeResource = NULL;
pfnRtlDeleteResource rtlDeleteResource = NULL;
pfnRtlDumpResource rtlDumpResource = NULL;
pfnRtlReleaseResource rtlReleaseResource = NULL;
pfnRtlAcquireResourceExclusive rtlAcquireResourceExclusive = NULL;
pfnRtlAcquireResourceShared rtlAcquireResourceShared = NULL;
#pragma endregion
#pragma endregion

/** Gets the address of a procedure in a module

	\tparam Fn Implied type of the function pointer to retrieved
	\param hMod Module to lookup the function in
	\param [out] fn Address of the function
	\param name Function to lookup
*/
template<class Fn>
void GPA(HMODULE hMod, Fn& fn, LPCSTR name)
{
	fn = reinterpret_cast<Fn>(GetProcAddress(hMod, name));
}

/** Initialises a resource 

	If this is the first object initialised, the function
	pointers to the Rtl*Resource functions in ntdll.dll are 
	retreived before initialisation.
*/
RtlResource::RtlResource()
{
	if(!rtlInitializeResource)
	{
		HMODULE hNt = GetModuleHandle(L"ntdll.dll");
		GPA(hNt, rtlDeleteResource, "RtlDeleteResource");
		GPA(hNt, rtlDumpResource, "RtlDumpResource");
		GPA(hNt, rtlReleaseResource, "RtlReleaseResource");
		GPA(hNt, rtlAcquireResourceExclusive, "RtlAcquireResourceExclusive");
		GPA(hNt, rtlAcquireResourceShared, "RtlAcquireResourceShared");
		GPA(hNt, rtlInitializeResource, "RtlInitializeResource");
	}
	rtlInitializeResource(&resource);
}

/** Frees the memory and handles associated with the resource */
RtlResource::~RtlResource()
{
	rtlDeleteResource(&resource);
}

/** Acquires the resource for exclusive access

	Exclusive access cannot be requested if the thread already has shared access,
	this will lead to deadlock unless waitForAccess is FALSE

	\param waitForAccess Specifies whether to block if access can't be immediately granted
	\returns Non-zero if access was granted, zero if not
*/
BOOLEAN RtlResource::AcquireExclusive(BOOLEAN waitForAccess)
{
	return rtlAcquireResourceExclusive(&resource, waitForAccess);
}

/** Acquires the resource for shared access

	Shared access can be requested if the thread already has exclusive access,
	the thread however retains the exclusive access state

	\param waitForAccess Specifies whether to block if access can't be immediately granted
	\returns Non-zero if access was granted, zero if not
*/
BOOLEAN RtlResource::AcquireShared(BOOLEAN waitForAccess)
{
	return rtlAcquireResourceShared(&resource, waitForAccess);
}

/** Outputs details about the current state of the resource to the debug output

	The display is of the format:\n
	<tt>Resource @ \<address of resource\>\n
	NumberOfWaitingShared = \<number\>\n
	NumberOfWaitingExclusive = \<number\>\n
	NumberOfActive = \<number\></tt>
*/
void RtlResource::OutputDebugText() const
{
	rtlDumpResource(&resource);
}

/** Undoes the previous acquire operation whether shared or exclusive */
void RtlResource::Release()
{
	rtlReleaseResource(&resource);
}

/** \endcond */
