/** \file rtlresource.h
	\brief Definitions for functions and objects using Microsoft fat read/write lock
	\cond DEV
*/

#ifndef AUDRECORD_RTLRESOURCE_H
#define AUDRECORD_RTLRESOURCE_H

#pragma once

/** Microsoft's read/write lock that is downgradable and recursive.

    Can be acquired for shared access if you have exclusive access
	but you retain the exclusive state.
*/
typedef struct RTL_RESOURCE
{
	CRITICAL_SECTION CSection;
	HANDLE hSharedSemaphore;
	ULONG sharedExclusive;
	HANDLE hExclusiveSemaphore;
	ULONG exclusiveWaiters;
	ULONG numberOfWaiters;
	HANDLE hOwnerThread;
	ULONG flags;
	PRTL_RESOURCE_DEBUG DebugInfo;
} *PRTL_RESOURCE;

/** Very thin RAII wrapper around the fat read/writer lock */
class RtlResource
{
private:
	/** The lock being managed */
	RTL_RESOURCE resource;
	/** piracy prevention, to stop copying */
	RtlResource(const RtlResource&);
	void operator=(const RtlResource&);
public:
	RtlResource();
	~RtlResource();
	BOOLEAN AcquireShared(BOOLEAN waitForAccess);
	BOOLEAN AcquireExclusive(BOOLEAN waitForAccess);
	void Release();
	void OutputDebugText() const;
};

/** A RAII object that acquires an RtlResource for shared access */
struct SharedRtlResourceLock
{
private:
	/** Resource to acquire and release */
	RtlResource& r;
	SharedRtlResourceLock(const SharedRtlResourceLock&);
	void operator=(const SharedRtlResourceLock&);
public:
	/** Initialises the lock member and acquires it for shared access
		
		\param r The lock to acquire and subsequently release
		\param wait Whether to block for shared access if it cannot be granted immediately
	*/
	SharedRtlResourceLock(RtlResource& r, BOOLEAN wait = TRUE)
		: r(r)
	{
		r.AcquireShared(wait);
	}

	/** Releases the shared access lock */
	~SharedRtlResourceLock()
	{
		r.Release();
	}
};

/** A RAII object that acquires an RtlResource for exclusive access */
struct ExclusiveRtlResourceLock
{
private:
	/** Resource to acquire and release */
	RtlResource& r;
	ExclusiveRtlResourceLock(const ExclusiveRtlResourceLock&);
	void operator=(const ExclusiveRtlResourceLock&);
public:
	/** Initialises the lock member and acquires it for exclusive access
		
		\param r The lock to acquire and subsequently release
		\param wait Whether to block for exclusive access if it cannot be granted immediately
	*/
	ExclusiveRtlResourceLock(RtlResource& r, BOOLEAN wait = TRUE)
		: r(r)
	{
		r.AcquireExclusive(wait);
	}

	/** Releases the exclusive access lock */
	~ExclusiveRtlResourceLock()
	{
		r.Release();
	}
};

/** \endcond */
#endif
