/** \file scoped.h
	\brief RAII guards arounds Windows resources, arrays and COM objects
	\cond DEV
*/
#ifndef SCOPED_OBJECTS_H
#define SCOPED_OBJECTS_H

#pragma once

/** Wrapper around a raw array

	\tparam Type The types of objects that'll be stored in the array
*/
template<class Type>
class ScopedArray
{
private:
	/** Pointer to the array that the object owns */
	Type* data;

public:
	/** Construct an object with a preexisting array
	
		\param data A pointer that the object takes ownership of 
	*/
	ScopedArray(Type* data) : data(data)
	{}

	/** Construct an object by creating a new sized array

		\param elements The size of the new array
	*/
	ScopedArray(UINT elements) : data(new Type[elements])
	{}

	/** Construct an object to an uninitialized state.

		Init() can be used to initialise the array
	*/
	ScopedArray() : data(NULL)
	{}

	/** Frees the array memory using delete */
	~ScopedArray()
	{		
		delete [] data;
	}

	/** Retrieves the array pointer
	
		The function asserts that the object is initialised before
		returning the pointer

		\returns The objects data pointer
	*/
	Type* get() const
	{		
		ASSERT(data && "Array not initialized");
		return data;
	}

	/** Retrieves a pointer to the array pointer
	
		The function asserts that the object has not already
		been initialised before	returning the pointer.
		Allowing this can lead to memory leaks

		\returns A pointer to the objects data pointer
	*/
	Type** operator&()
	{		
		ASSERT(!data && "Array already initialized");
		return &data;
	}

	/** Retrieves the array pointer
	
		The function asserts that the object is initialised before
		returning the pointer

		\returns The objects data pointer
	*/
	Type* operator*() const
	{		
		return get();
	}

	/** Initialises an uninitialised object

		The function asserts that the object hasn't been initialised

		\param newData A pointer that the object takes ownership of.
		Must have been returned by operator new
	*/
	void Init(Type* newData)
	{		
		ASSERT(!data && "Array already initialized");
		data = newData;
	}

	/** Releases the object's ownership of the stored pointer

		Turns the object back to an uninitialised state. The caller is then
		responsible for deleting the array when it is no longer required
		
		\returns The stored array pointer
	*/
	Type* Release()
	{
		Type* pTemp = data;
		data = NULL;
		return pTemp;
	}

	/** Swaps the pointers stored in two disperate objects

		\param other The object to swap this objects pointer with
	*/
	void Swap(ScopedArray& other)
	{
		Type* pTemp = other.data;
		other.data = data;
		data = pTemp;
	}
};

/** Wrapper around a Windows resource such as HANDLE

	\tparam Type The types of objects that'll be stored and released. Defaults to HANDLE
	\tparam Ret Return type of the function that'll release said resource. Defaults to BOOL
*/
template<class Type = HANDLE, class Ret = BOOL>
class WinType
{
private:
	/** Function pointer type of the releasing function */
	typedef Ret (WINAPI*ReleaseProc)(Type);
	/** Safe bool idiom member function typedef - the Swap() method */
	typedef void (WinType:: * bool_type)(WinType&);

	/** The resource to free */
	Type data;
	/** The function to call that'll free the resource */
	ReleaseProc lpfnReleaser;

public:
	/** Constructs the object with a resource and a releaser
	
		\param data The resource to take ownership of
		\param lpfnReleaser The function used to free the resource on destruction
	*/
	WinType(const Type& data, ReleaseProc lpfnReleaser)
		: data(data),
		lpfnReleaser(lpfnReleaser)
	{}

	/** Constructs an object using the specified resource freeing function

		The resource is left uninitialised. This constructor is primarily for use
		with operator& and passing the uninitialised resource to the creation function as
		a pointer-pointer parameter. E.g.

		// create with the releaser function
		WinType<PBYTE, NET_API_STATUS> memWrap(&NetApiBufferFree);
		// initialise the wrapped resource indirectly
		NetGroupEnum(NULL, 1, &memWrap, ...);
	*/
	WinType(ReleaseProc lpfnReleaser)
		: data(),
		lpfnReleaser(lpfnReleaser)
	{}

	/** Constructs an uninitialised object
	
		Init must be called to initialise the object at a later time
	*/
	WinType()
		: data(),
		lpfnReleaser(NULL)
	{}

	/** Frees the resource if the object has been initialised*/
	~WinType()
	{		
		if(lpfnReleaser)
		{
			lpfnReleaser(data);
		}
	}

	/** Retrieves the wrapped resource
	
		The function asserts that the object is initialised before
		returning the pointer

		\returns The objects resource
	*/
	Type get() const
	{
		ASSERT(lpfnReleaser && "Object not initialised");
		return data;
	}

	/** Retrieves the wrapped resource
	
		Syntactic sugar for the get() function

		\returns The resource
	*/
	Type operator*() const
	{		
		return get();
	}

	/** Exposes the address of the wrapped, yet uninitialised resource 

		This operator is solely intended for use with the function pointer only constructor, 
		whose use is asserted in the function.

		\returns A pointer to an uninitialised resource
	*/
	Type* operator&()
	{		
		ASSERT((data == Type()) && lpfnReleaser && "Object already initialised");
		return &data;
	}

	/** Safe bool idiom initialised check

		\returns NULL if the object is in an uninitialised state, non-zero otherwise
	*/
	operator bool_type() const
	{
		return lpfnReleaser ? &WinType::Swap : NULL;
	}

	/** Initialises an uninitialised object

		Also asserts that the object hasn't been initialised

		\param newData A resource that the object takes ownership of.
		\param lpfnNewReleaser Function to use to release newData
	*/
	void Init(const Type& newData, ReleaseProc lpfnNewReleaser)
	{
		ASSERT((!lpfnReleaser) && "Object already initialized");
		data = newData;
		lpfnReleaser = lpfnNewReleaser;
	}

	/** Releases the object's ownership of the stored resource

		Returns the object back to an uninitialised state. The caller is then
		responsible for releasing the resource when it is no longer required.
		
		\returns The stored resource
	*/
	Type Release()
	{		
		lpfnReleaser = NULL;
		Type temp = data;
		data = Type();
		return temp;
	}

	/** Swaps the resources and functions stored in two disperate objects

		\param other The object to swap this objects internals with
	*/
	void Swap(WinType& other)
	{
		Type temp = other.data;
		ReleaseProc tempProc = other.lpfnReleaser;
		other.data = data;
		other.lpfnReleaser = lpfnReleaser;
		data = temp;
		lpfnReleaser = tempProc;
	}
};

/** Scoped wrapper around a COM object

	\tparam Type The interface type that'll be stored in the object
*/
template<class Type>
class ComType
{
private:
	/** The interface pointer that will be Released on destruction */
	Type* data;

	/** Safe bool idiom member function typedef - the IUnknown::Release method */
	typedef ULONG (STDMETHODCALLTYPE Type::IUnknown:: * bool_type)();

public:
	/** Construct an object with a preexisting interface pointer

		The function assumes ownership of the pointer and does not call
		AddRef on it.
	
		\param data A pointer that the object takes ownership of 
	*/
	ComType(Type* data)
		: data(data)
	{
		// check that this is actually a com interface pointer
		// and not anything else
		C_ASSERT((__is_base_of(IUnknown, Type)));
	}

	/** Copies the pointer stored in another object and calls AddRef

		\param other The object to copy the pointer from
	*/
	ComType(const ComType& other)
		: data(other.data)
	{
		if(data)
		{
			data->AddRef();
		}
	}

	/** Leaves the object in an uninitialised state
	*/
	ComType()
		: data(NULL)
	{}

	/** Calls Release on the stored interface pointer if there is one
	*/
	~ComType()
	{		
		if(data)
		{
			data->Release();
		}
	}

	/** Makes the contained pointer equal to that contained in another object

		\param other The object and indirectly the pointer to make this object the same as
	*/
	void operator=(const ComType& other)
	{
		ComType temp(other);
		Swap(temp);
	}

	/** Retrieves the stored pointer
	
		The function asserts that the object is initialised before
		returning the pointer

		\returns The objects data pointer
	*/
	Type* get() const
	{
		ASSERT(data && "No object set");
		return data;
	}

	/** Retrieves the stored pointer
	
		This operator is just syntactic sugar around the get() method.

		\returns The objects data pointer
	*/
	Type* operator*() const
	{		
		return get();
	}

	/** Retrieves a pointer to the interface pointer
	
		The function asserts that the object has not already
		been initialised before	returning the pointer.
		Allowing this can lead to memory leaks

		\returns A pointer to the stored interface pointer
	*/
	Type** operator&()
	{		
		ASSERT(!data && "Object already initialized");
		return &data;
	}

	/** Retrieves the stored pointer

		Equivalent to get(), this operator allows the object
		to be used as a pointer
		\returns The stored interface pointer
	*/
	Type* operator->()
	{
		return get();
	}

	/** Safe bool idiom initialised check

		\returns NULL if the object is in an uninitialised state, non-zero otherwise
	*/
	operator bool_type() const
	{
		return data ? &Type::Release : 0;
	}

	/** Initialises an uninitialised object

		The function asserts that the object hasn't been initialised

		\param newData An interface pointer that the object takes ownership of
	*/
	void Init(Type* newData)
	{		
		ASSERT(!data && "Object already initialized");
		data = newData;
	}

	/** Releases the object's ownership of the stored pointer

		Turns the object back to an uninitialised state. The caller is then
		responsible for calling Release() on the interface when it is no longer required.
		Note: this function DOES NOT call Release on the stored interface pointer
		and is not equivalent to calling it.
		
		\returns The stored pointer
	*/
	Type* Release()
	{
		Type* temp = data;
		data = NULL;
		return temp;
	}

	/** Swaps the pointers stored in two disperate objects

		\param other The object to swap with
	*/
	void Swap(ComType& other)
	{
		Type* pTemp = other.data;
		other.data = data;
		data = pTemp;
	}
};

/** RAII wrapper around acquiring and releasing shared access
	on a Slim Read-Writer lock
*/
struct SharedSRWLock
{
	/** The lock to acquire and release */
	SRWLOCK* pLock;

	/** Initialises the object and acquires shared access to the lock

		The function will block until shared access can be achieved
		\param pLock The lock to take
	*/
	SharedSRWLock(SRWLOCK* pLock)
		: pLock(pLock)
	{
		AcquireSRWLockShared(pLock);
	}

	/** Releases the lock acquired in the constructor */
	~SharedSRWLock()
	{
		ReleaseSRWLockShared(pLock);
	}
};

/** RAII wrapper around acquiring and releasing exclusive access
	on a Slim Read-Writer lock
*/
struct ExclusiveSRWLock
{
	/** The lock to acquire and release */
	SRWLOCK* pLock;

	/** Initialises the object and acquires exclusive access to the lock

		The function will block until exclusive access can be achieved
		\param pLock The lock to take
	*/
	ExclusiveSRWLock(SRWLOCK* pLock)
		: pLock(pLock)
	{
		AcquireSRWLockExclusive(pLock);
	}

	/** Releases the lock acquired in the constructor */
	~ExclusiveSRWLock()
	{
		ReleaseSRWLockExclusive(pLock);
	}
};

/** \endcond */

#endif
